import datetime
from textwrap import shorten
from typing import Any, Dict, Optional, Sequence, Union

import aiohttp
import feedparser
from pydantic import ConfigDict, ValidationError

from avtdl.core.interfaces import MAX_REPR_LEN, Record, TextRecord
from avtdl.core.monitors import BaseFeedMonitor, BaseFeedMonitorConfig, BaseFeedMonitorEntity
from avtdl.core.plugins import Plugins
from avtdl.core.utils import html_to_text, make_datetime


@Plugins.register('generic_rss', Plugins.kind.ASSOCIATED_RECORD)
class GenericRSSRecord(Record):
    """
    Represents RSS feed entry

    Might contain additional fields if they are present in the feed.
    """
    model_config = ConfigDict(extra='allow')

    uid: str
    """value that is unique for this entry of RSS feed"""
    url: str
    """"href" or "link" field value of this entry"""
    summary: str
    """"summary" or "description" field value of this entry"""
    author: str = ''
    """"author" field value. Might be empty"""
    title: str = ''
    """"title" field value. Might be empty"""
    published: datetime.datetime
    """"published" or "issued" field value of this entry"""

    def __str__(self):
        second_line = f'[{self.published}] {self.author}: {self.title}\n' if self.author or self.title else ''
        summary = html_to_text(self.summary)
        summary = shorten(summary, MAX_REPR_LEN * 5)
        return f'{self.url}\n{second_line}{summary}'

    def __repr__(self):
        title = shorten(self.title, MAX_REPR_LEN)
        return f'GenericRSSRecord(updated="{self.published.isoformat()}", url="{self.url}", title="{title}")'

    def get_uid(self) -> str:
        return self.uid


@Plugins.register('generic_rss', Plugins.kind.ACTOR_CONFIG)
class GenericRSSMonitorConfig(BaseFeedMonitorConfig):
    pass

@Plugins.register('generic_rss', Plugins.kind.ACTOR_ENTITY)
class GenericRSSMonitorEntity(BaseFeedMonitorEntity):
    pass

@Plugins.register('generic_rss', Plugins.kind.ACTOR)
class GenericRSSMonitor(BaseFeedMonitor):
    """
    RSS feed monitor

    Monitors RSS feed for new entries. Will attempt to adjust
    update interval based on HTTP response headers.

    Depending on a specific feed format, fields names and content
    might vary greatly. Commonly present standardized fields are
    `url`, `title` and `author`, though they might be empty
    in some feeds.

    Before defining a command to be executed for records of a newly
    added feed it is recommended to inspect the feed entity content by
    forwarding records in a file in JSON format using `to_file` plugin.

    Normally feeds have some kind of value to uniquely identify
    feed entries, but in case there is none, parser will attempt
    to create one by combining `link` and `title` or `summary` fields.
    """

    async def get_records(self, entity: BaseFeedMonitorEntity, session: aiohttp.ClientSession) -> Sequence[GenericRSSRecord]:
        raw_feed = await self._get_feed(entity, session)
        if raw_feed is None:
            return []
        records = self._parse_entries(raw_feed)
        return records

    async def _get_feed(self, entity: BaseFeedMonitorEntity, session: aiohttp.ClientSession) -> Optional[feedparser.FeedParserDict]:
        response = await self.request_raw(entity.url, entity, session)
        if response is None:
            return None
        raw_text = await response.text()

        # Charset detection used by aiohttp fails on RSS feed often enough,
        # and feedparser is capable of taking bytes as input and detecting
        # encoding itself. Since aiohttp.ClientResponse does not provide
        # a way to get response body as raw bytes multiple times, encode
        # text back to bytes instead.
        # Skip it for UTF-8 since it's most likely to be correct.
        text: Union[str, bytes] = raw_text
        if not response.get_encoding().lower() in ('utf-8', 'utf8'):
            self.logger.debug(f'[{entity.name}] detected encoding is {response.get_encoding()}, reverting to bytes')
            try:
                text = raw_text.encode(response.get_encoding())
            except Exception as e:
                self.logger.exception(f'[{entity.name}] failed to encode raw feed back to bytes: {e}')

        # Set "content-location" header to allow feedparser to resolve
        # relative links in feed entries. Header name must be
        # all lowercase because this is how it is checked in feedparser,
        # and it uses regular case-sensitive dict for headers internally
        response_headers = response.headers.copy()
        if 'Content-Location' in response.headers:
            response_headers['content-location'] = response_headers['Content-Location']
        else:
            response_headers['content-location'] = entity.url

        try:
            feed = feedparser.parse(text, response_headers=response_headers, resolve_relative_uris=True)
            if feed.get('entries') is not None:
                return feed
            else:
                from pprint import pformat

                self.logger.debug(f'feed for {entity.url} has no entries, probably broken:')
                self.logger.debug(pformat(feed))
                raise Exception(f'got broken feed while fetching {entity.url}')
        except Exception as e:
            self.logger.warning('Exception while updating rss feed: {}'.format(e))
            return None

    @staticmethod
    def _get_uid(entry: feedparser.FeedParserDict) -> str:
        # https://www.詹姆斯.com/blog/2006/08/rss-dup-detection
        guid = entry.get('guid') or entry.get('id')
        if guid is not None:
            return guid
        link = entry.get('link') or entry.get('href', '')
        title = entry.get('title', '')
        summary = entry.get('summary', '')
        if any([link, title, summary]):
            return link + (title or summary)
        return TextRecord(text=str(entry)).hash() #short way to write sha1(str(entry))

    @classmethod
    def _parse_entry(cls, entry: feedparser.FeedParserDict) -> GenericRSSRecord:
        # base properties with substitutions defined for feedparser.FeedParserDict
        # keymap = {
        #     'channel': 'feed',
        #     'items': 'entries',
        #     'guid': 'id',
        #     'date': 'updated',
        #     'date_parsed': 'updated_parsed',
        #     'description': ['summary', 'subtitle'],
        #     'description_detail': ['summary_detail', 'subtitle_detail'],
        #     'url': ['href'],
        #     'modified': 'updated',
        #     'modified_parsed': 'updated_parsed',
        #     'issued': 'published',
        #     'issued_parsed': 'published_parsed',
        #     'copyright': 'rights',
        #     'copyright_detail': 'rights_detail',
        #     'tagline': 'subtitle',
        #     'tagline_detail': 'subtitle_detail',
        # }

        parsed: Dict[str, Any] = {}

        parsed['uid'] = cls._get_uid(entry)
        parsed['url'] = entry.pop('link', '') or entry.pop('href', '') or entry.pop('url', '')
        parsed['summary'] = entry.pop('summary', '')
        parsed['author'] = entry.pop('author', '')
        parsed['title'] = entry.pop('title', '')

        parsed['published'] = make_datetime(entry.pop('published_parsed'))
        entry.pop('published')
        updated = entry.pop('updated_parsed', None)
        if updated is not None:
            parsed['updated'] = make_datetime(updated)
            entry.pop('updated', '')

        for key, value in entry.items():
            parsed[key] = value

        record = GenericRSSRecord(**parsed)
        return record

    def _parse_entries(self, feed: feedparser.FeedParserDict) -> Sequence[GenericRSSRecord]:
        records = []
        for entry in feed['entries']:
            try:
                record = self._parse_entry(entry)
            except KeyError as e:
                self.logger.warning(f'rss parser failed to construct record from rss entry {entry}: missing  necessarily field {e}')
                continue
            except ValidationError as e:
                self.logger.warning(f'rss parser failed to construct record from rss entry {entry}: {e}')
                continue
            records.append(record)
        records = records[::-1]  # records are ordered from new to old in the feed, reorder in chronological order
        return records



# Prepare request to retrieve json data from graphql endpoints
#
# - timeline (home and chronological)
# - user tweets (without and with replies)
# - user likes
# - single tweet
#
# - user id by screen name
import abc
import datetime
import json
import logging
import urllib.parse
from dataclasses import dataclass
from http.cookiejar import CookieJar
from typing import Any, Dict, Optional, Union

import aiohttp
from multidict import CIMultiDictProxy

from avtdl.core import utils
from avtdl.core.utils import RateLimit, find_all, find_one, get_retry_after

USER_FEATURES = '{"hidden_profile_likes_enabled":true,"hidden_profile_subscriptions_enabled":true,"rweb_tipjar_consumption_enabled":true,"responsive_web_graphql_exclude_directive_enabled":true,"verified_phone_label_enabled":false,"subscriptions_verification_info_is_identity_verified_enabled":true,"subscriptions_verification_info_verified_since_enabled":true,"highlights_tweets_tab_ui_enabled":true,"responsive_web_twitter_article_notes_tab_enabled":true,"creator_subscriptions_tweet_preview_api_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"responsive_web_graphql_timeline_navigation_enabled":true}'
TWEETS_FEATURES = '{"rweb_tipjar_consumption_enabled":true,"responsive_web_graphql_exclude_directive_enabled":true,"verified_phone_label_enabled":false,"creator_subscriptions_tweet_preview_api_enabled":true,"responsive_web_graphql_timeline_navigation_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"communities_web_enable_tweet_community_results_fetch":true,"c9s_tweet_anatomy_moderator_badge_enabled":true,"tweetypie_unmention_optimization_enabled":true,"responsive_web_edit_tweet_api_enabled":true,"graphql_is_translatable_rweb_tweet_is_translatable_enabled":true,"view_counts_everywhere_api_enabled":true,"longform_notetweets_consumption_enabled":true,"responsive_web_twitter_article_tweet_consumption_enabled":true,"tweet_awards_web_tipping_enabled":false,"creator_subscriptions_quote_tweet_preview_enabled":false,"freedom_of_speech_not_reach_fetch_enabled":true,"standardized_nudges_misinfo":true,"tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled":true,"rweb_video_timestamps_enabled":true,"longform_notetweets_rich_text_read_enabled":true,"longform_notetweets_inline_media_enabled":true,"responsive_web_enhance_cards_enabled":false}'
SPACE_FEATURES = '{"spaces_2022_h2_spaces_communities":true,"spaces_2022_h2_clipping":true,"creator_subscriptions_tweet_preview_api_enabled":true,"rweb_tipjar_consumption_enabled":true,"responsive_web_graphql_exclude_directive_enabled":true,"verified_phone_label_enabled":false,"communities_web_enable_tweet_community_results_fetch":true,"c9s_tweet_anatomy_moderator_badge_enabled":true,"articles_preview_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"tweetypie_unmention_optimization_enabled":true,"responsive_web_edit_tweet_api_enabled":true,"graphql_is_translatable_rweb_tweet_is_translatable_enabled":true,"view_counts_everywhere_api_enabled":true,"longform_notetweets_consumption_enabled":true,"responsive_web_twitter_article_tweet_consumption_enabled":true,"tweet_awards_web_tipping_enabled":false,"creator_subscriptions_quote_tweet_preview_enabled":false,"freedom_of_speech_not_reach_fetch_enabled":true,"standardized_nudges_misinfo":true,"tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled":true,"rweb_video_timestamps_enabled":true,"longform_notetweets_rich_text_read_enabled":true,"longform_notetweets_inline_media_enabled":true,"responsive_web_graphql_timeline_navigation_enabled":true,"responsive_web_enhance_cards_enabled":false}'


class EndpointUrl:
    TIMELINE = 'https://twitter.com/i/api/graphql/uPv755D929tshj6KsxkSZg/HomeTimeline'
    LATEST_TIMELINE = 'https://twitter.com/i/api/graphql/70b_oNkcK9IEN13WNZv8xA/HomeLatestTimeline'
    USER_TWEETS = 'https://twitter.com/i/api/graphql/piUHOePH_uDdwbD9GkquJA/UserTweets'
    USER_TWEETS_AND_REPLIES = 'https://twitter.com/i/api/graphql/KJiZSYLD2ijyHRBmgddo8Q/UserTweetsAndReplies'
    USER_LIKES = 'https://twitter.com/i/api/graphql/W42Y54_EmIjbTEdg9mGLDQ/Likes'
    TWEET_DETAIL = 'https://twitter.com/i/api/graphql/F45teiuFI9MDxaS9UYKv-g/TweetDetail'
    USER_BY_SCREEN_NAME = 'https://twitter.com/i/api/graphql/qW5u-DAuXpMEG0zA1F7UGQ/UserByScreenName'
    AUDIOSPACE_BY_ID = 'https://twitter.com/i/api/graphql/d03OdorPdZ_sH9V3D1_yWQ/AudioSpaceById'


@dataclass
class RequestDetails:
    url: str
    params: Dict[str, Any]
    headers: Dict[str, Any]
    cookies: CookieJar

    def with_base_url(self, base_url: str) -> 'RequestDetails':
        new_url = replace_url_host(self.url, base_url)
        return RequestDetails(url=new_url, params=self.params, headers=self.headers, cookies=self.cookies)


def replace_url_host(url: str, new_host: str) -> str:
    parsed_host = urllib.parse.urlparse(new_host)
    parsed_url = urllib.parse.urlparse(url)
    scheme = parsed_url.scheme or parsed_host.scheme or 'https://'
    new_url = urllib.parse.urlunparse((scheme, parsed_host.netloc, parsed_url.path, parsed_url.params, parsed_url.query, parsed_url.fragment))
    return new_url


def get_netloc(host: str) -> str:
    return urllib.parse.urlparse(host).netloc


def get_cookie_value(jar: Union[CookieJar, aiohttp.CookieJar], name: str) -> Optional[str]:
    if isinstance(jar, CookieJar):
        found = [x for x in jar if x.name == name]
    else:
        found = [x for x in jar if x.key == name]

    if not found:
        return None
    return found[0].value


def get_auth_headers(cookies) -> dict[str, Any]:
    ct0 = get_cookie_value(cookies, 'ct0') or ''
    THE_API_KEY = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs=1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'
    headers = {
        'authorization': THE_API_KEY,
        'x-csrf-token': ct0
    }
    return headers


class TwitterRateLimit(RateLimit):

    def _submit_headers(self, headers: Union[Dict[str, str], CIMultiDictProxy[str]], logger: logging.Logger):
        try:
            self.limit_total = int(headers.get('x-rate-limit-limit', -1))
            self.limit_remaining = int(headers.get('x-rate-limit-remaining', -1))
            self.reset_at = int(headers.get('x-rate-limit-reset', -1))
        except ValueError:
            logger.warning(f'[{self.name}] error parsing rate limit headers: "{headers}"')
        else:
            logger.debug(f'[{self.name}] rate limit {self.limit_remaining}/{self.limit_total}, resets after {datetime.timedelta(seconds=self.reset_after)}')


class TwitterEndpoint(abc.ABC):
    """
    Superclass providing utility methods for concrete Endpoints

    Concrete Endpoints must implement prepare() method taking arbitrary
    arguments, that returns RequestDetails instance.

    They might use methods provided by this class for convenience,
    however, implementing them is not required.
    """
    FEATURES = TWEETS_FEATURES
    URL = 'https://twitter.com/...'

    _rate_limit: Optional[TwitterRateLimit] = None

    @classmethod
    def rate_limit(cls) -> TwitterRateLimit:
        if cls._rate_limit is None:
            cls._rate_limit = TwitterRateLimit(cls.__name__)
        return cls._rate_limit

    @staticmethod
    def get_base_variables(has_continuation: bool = False):
        raise NotImplementedError

    @classmethod
    def get_variables(cls, continuation: Optional[str] = None, count: int = 20, user_id: Optional[str] = None) -> str:
        variables = cls.get_base_variables(continuation is not None)
        if continuation is not None:
            variables['cursor'] = continuation
        variables['count'] = count
        if user_id is not None:
            variables['userId'] = user_id
        variables_text = json.dumps(variables)
        return variables_text

    @classmethod
    def prepare_for(cls, host, cookies, variables: str) -> RequestDetails:
        url = replace_url_host(cls.URL, host)

        params = {}
        params['variables'] = variables
        params['features'] = cls.FEATURES

        headers = get_auth_headers(cookies)

        details = RequestDetails(url=url, params=params, headers=headers, cookies=cookies)
        return details

    @classmethod
    @abc.abstractmethod
    def prepare(cls, *args, **kwargs) -> RequestDetails:
        """Prepare a RequestDetails object based on passed arguments"""

    @classmethod
    async def request_raw(cls, logger: logging.Logger, session: aiohttp.ClientSession, *args, **kwargs) -> Optional[aiohttp.ClientResponse]:
        r = cls.prepare(*args, **kwargs)
        if r is None:
            return None
        async with cls.rate_limit() as rate_limit:
            try:
                response = await utils.request_raw(r.url, session, logger, params=r.params, headers=r.headers, retry_times=0, raise_errors=True)
            except Exception as e:
                if isinstance(e, aiohttp.ClientResponseError):
                    rate_limit.submit_headers(e.headers, logger)
                    logger.debug(f' got code {e.status} ({e.message}) while fetching {r.url}: {e}')
                else:
                    logger.debug(f'error while fetching {r.url}: {e}')
                return None
            assert response is not None, 'request_raw() returned None despite raise_errors=True'
            rate_limit.submit_headers(response.headers, logger)
            return response

    @classmethod
    async def request(cls, logger: logging.Logger, session: aiohttp.ClientSession, *args, **kwargs) -> Optional[str]:
        response = await cls.request_raw(logger, session, *args, **kwargs)
        if response is None:
            return None
        return await response.text()


class UserIDEndpoint(TwitterEndpoint):
    URL = EndpointUrl.USER_BY_SCREEN_NAME
    FEATURES = USER_FEATURES

    @classmethod
    def prepare(cls, host: str, cookies, user_handle: str) -> RequestDetails:
        user_handle = user_handle.strip('@/')
        variables = {'screen_name': user_handle, 'withSafetyModeUserFields': True}
        variables_text = json.dumps(variables)
        return cls.prepare_for(host, cookies, variables_text)

    @staticmethod
    def get_user_id(data: dict) -> Optional[str]:
        return find_one(data, '$.data.user.result.rest_id')


class TweetDetailEndpoint(TwitterEndpoint):
    URL = EndpointUrl.TWEET_DETAIL

    @classmethod
    def prepare(cls, host: str, cookies, tweet_id: str, continuation: Optional[str] = None) -> RequestDetails:
        variables = {'focalTweetId': tweet_id, 'with_rux_injections': False, 'includePromotedContent': True,
                     'withCommunity': True, 'withQuickPromoteEligibilityTweetFields': True, 'withBirdwatchNotes': True,
                     'withVoice': True, 'withV2Timeline': True}
        if continuation:
            variables.update({'referrer': 'tweet', 'cursor': continuation})
        variables_text = json.dumps(variables)
        return cls.prepare_for(host, cookies, variables_text)


class AudioSpaceEndpoint(TwitterEndpoint):
    URL = EndpointUrl.AUDIOSPACE_BY_ID
    FEATURES = SPACE_FEATURES

    @classmethod
    def prepare(cls, host, cookies, space_id: str) -> RequestDetails:
        variables = {'id': space_id, 'isMetatagsQuery': False, 'withReplays': True, 'withListeners': True}
        variables_text = json.dumps(variables)
        return cls.prepare_for(host, cookies, variables_text)


class LiveStreamEndpoint(TwitterEndpoint):
    """Endpoint for retrieving HLS playlist for AudioSpace"""
    url = 'https://twitter.com/i/api/1.1/live_video_stream/status/{}'

    @classmethod
    def prepare(cls, host: str, cookies, media_key: str) -> RequestDetails:
        url = cls.url.format(media_key)
        url = replace_url_host(url, host)
        params = {'client': 'web', 'use_syndication_guest_id': 'false', 'cookie_set_host': get_netloc(host)}
        headers = get_auth_headers(cookies)

        details = RequestDetails(url=url, params=params, headers=headers, cookies=cookies)
        return details


class LatestTimelineEndpoint(TwitterEndpoint):
    URL = EndpointUrl.LATEST_TIMELINE

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        if not is_continuation:
            request_context = 'launch'
        else:
            request_context = 'ptr'
        variables = {'includePromotedContent': True, 'latestControlAvailable': True, 'requestContext': request_context}
        return variables

    @classmethod
    def prepare(cls, host: str, cookies, continuation: Optional[str] = None, count: int = 20) -> RequestDetails:
        variables = cls.get_variables(continuation=continuation, count=count)
        return cls.prepare_for(host, cookies, variables)


class TimelineEndpoint(LatestTimelineEndpoint):
    URL = EndpointUrl.TIMELINE

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        return {"includePromotedContent": True, "latestControlAvailable": True, "withCommunity": True}


class UserTweetsEndpoint(TwitterEndpoint):
    URL = EndpointUrl.USER_TWEETS

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        return {'includePromotedContent': True, 'withQuickPromoteEligibilityTweetFields': True, 'withVoice': True, 'withV2Timeline': True}

    @classmethod
    def prepare(cls, host: str, cookies, user_id: str, continuation: Optional[str] = None, count: int = 20) -> RequestDetails:
        variables = cls.get_variables(user_id=user_id, count=count, continuation=continuation)
        return cls.prepare_for(host, cookies, variables)


class UserTweetsRepliesEndpoint(UserTweetsEndpoint):
    URL = EndpointUrl.USER_TWEETS_AND_REPLIES

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        return {'includePromotedContent': True, 'withCommunity': True, 'withVoice': True, 'withV2Timeline': True}


class UserLikesEndpoint(UserTweetsEndpoint):
    URL = EndpointUrl.USER_LIKES

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        return {'includePromotedContent': False, 'withClientEventToken': False, 'withBirdwatchNotes': False, 'withVoice': True, 'withV2Timeline': True}


def get_rate_limit_delay(headers: Union[Dict[str, str], CIMultiDictProxy[str]], logger: Optional[logging.Logger] = None) -> int:
    logger = logger or logging.getLogger().getChild('twitter_endpoints')
    retry_after = get_retry_after(headers)
    if retry_after is not None:
        return retry_after
    try:
        limit_total = int(headers.get('x-rate-limit-limit', -1))
        limit_remaining = int(headers.get('x-rate-limit-remaining', -1))
        reset_at = int(headers.get('x-rate-limit-reset', -1))
    except ValueError as e:
        logger.debug(f'error parsing limit headers: "{headers}"')
        return 0
    now = int(datetime.datetime.now().timestamp())
    reset_after = max(0, reset_at - now)
    logger.debug(f'rate limit {limit_remaining}/{limit_total}, resets after {reset_after} (at {reset_at})')
    if limit_remaining <= 1:
        return reset_after + 1
    return 0


def get_continuation(data: dict) -> Optional[str]:
    entries = find_all(data, '$..instructions..entries..content,itemContent')
    continuation = entries[-1]['value']
    return continuation



from pathlib import Path
from typing import Dict, List, Tuple, Union

import pytest
from pydantic import BaseModel

from avtdl.avtdl import parse_config
from avtdl.core.config import config_sancheck
from avtdl.core.interfaces import Actor, Record, TextRecord
from avtdl.core.loggers import set_logging_format, silence_library_loggers
from avtdl.core.yaml import yaml_load
from avtdl.plugins.utils.utils import Consumer, Producer


class Sender(BaseModel):
    actor: str = 'utils.producer'
    entity: str
    records: List[Union[str, Record]]

    def model_post_init(self, __context):
        updated_records = []
        for item in self.records:
            if isinstance(item, str):
                record = TextRecord(text=item)
            else:
                record = item
            updated_records.append(record)
        self.records = updated_records


class Receiver(BaseModel):
    actor: str = 'utils.consumer'
    entity: str
    expected_history: List[Union[str, Record]]

    def model_post_init(self, __context):
        updated_expected_history = []
        for item in self.expected_history:
            if isinstance(item, str):
                record = TextRecord(text=item)
            else:
                record = item
            updated_expected_history.append(record)
        self.expected_history = updated_expected_history


class _Testcases(BaseModel):
    senders: List[Sender]
    receivers: List[Receiver]

    @classmethod
    def load(cls, data: dict) -> Tuple[List[Sender], List[Receiver]]:
        testcases = cls.model_validate(data)
        return testcases.senders, testcases.receivers


async def run(config: str):
    silence_library_loggers()
    set_logging_format('WARNING')

    conf: dict = yaml_load(config)
    senders, receivers = _Testcases.load(conf.pop('testcases'))
    _, actors, chains = parse_config(conf)
    config_sancheck(actors, chains)

    send_records(actors, senders)
    check_received(actors, receivers)


def send_records(actors: Dict[str, Actor], senders: List[Sender]):
    for sender in senders:
        actor = actors[sender.actor]
        entity = actor.entities[sender.entity]
        if not isinstance(actor, Producer):
            raise Exception(f'incorrect sender: expected TestMonitor, got {actor.__class__}')
        for record in sender.records:
            assert isinstance(record, Record), f'post_init hook failed for {sender}'
            actor.produce(entity.name, record)


def check_received(actors: Dict[str, Actor], receivers: List[Receiver]):
    for receiver in receivers:
        actor = actors[receiver.actor]
        entity = actor.entities[receiver.entity]
        if not isinstance(actor, Consumer):
            raise Exception(f'incorrect receiver: expected TestAction, got {actor.__class__}')
        assert all(isinstance(record, Record) for record in receiver.expected_history), f'post_init hook failed for {receiver}'

        actual_history = actor.history[entity.name]
        assert actual_history == receiver.expected_history


def load_testcases() -> Dict[str, str]:
    testcases: Dict[str, str] = {}
    files = Path(__file__).parent.glob('*.yml')
    for file in files:
        with open(file, 'rt', encoding='utf8') as fp:
            testcases[file.name] = fp.read()
    return testcases


testcases = load_testcases()


@pytest.mark.asyncio
@pytest.mark.parametrize('extended_config', list(testcases.values()), ids=testcases.keys())
async def test_chains(extended_config: str):
    await run(extended_config)



import datetime
import json
from json import JSONDecodeError
from textwrap import shorten
from typing import Optional, Sequence

import aiohttp
import pydantic
from pydantic import Field, PositiveFloat

from avtdl.core.config import Plugins
from avtdl.core.interfaces import ActorConfig, MAX_REPR_LEN, Record
from avtdl.core.monitors import HttpTaskMonitor, HttpTaskMonitorEntity
from avtdl.core.utils import Fmt, parse_timestamp_ms


@Plugins.register('fc2', Plugins.kind.ASSOCIATED_RECORD)
class FC2Record(Record):
    """Represents event of a stream going live on FC2"""
    name: str = ''
    """name of the config entity for this user"""
    url: str
    """url of the user stream"""
    user_id: str
    """unique for the given user/channel part of the stream url"""
    title: str
    """stream title"""
    info: str
    """stream description"""
    start: Optional[datetime.datetime]
    """time of the stream start"""
    start_timestamp: str
    """UNIX timestamp of the stream start"""
    avatar_url: str
    """link to the user's avatar"""
    login_only: bool
    """whether logging in is required to view current livestream"""

    def __str__(self):
        since = '\nsince ' + Fmt.date(self.start) if self.start else ''
        return f'{self.url}\n{self.title}{since}'

    def __repr__(self):
        title = shorten(self.title, MAX_REPR_LEN)
        return f'FC2Record(user_id={self.user_id}, start={self.start_timestamp}, title={title})'

    def discord_embed(self) -> dict:
        return {
            'title': self.title,
            'description': self.url,
            'color': None,
            'author': {'name': self.name, 'url': self.url, 'icon_url': self.avatar_url},
            'timestamp': self.start.isoformat() if self.start else None,
            'footer': {'text': self.info},
            'fields': []
        }


@Plugins.register('fc2', Plugins.kind.ACTOR_CONFIG)
class FC2MonitorConfig(ActorConfig):
    pass


@Plugins.register('fc2', Plugins.kind.ACTOR_ENTITY)
class FC2MonitorEntity(HttpTaskMonitorEntity):
    user_id: str
    """user id, numeric part at the end of livestream url"""
    update_interval: PositiveFloat = 120
    """how often the monitored channel will be checked, in seconds"""
    adjust_update_interval: bool = Field(exclude=True, default=True)
    """does nothing since fc2 does not use caching headers on the endpoint used to check live status"""
    latest_live_start: str = Field(exclude=True, default='')
    """internal variable to persist state between updates. Used to distinguish between different livestreams of the user"""


@Plugins.register('fc2', Plugins.kind.ACTOR)
class FC2Monitor(HttpTaskMonitor):
    """
    Monitor for live.fc2.com

    Monitors fc2.com user with given id, produces a record when it goes live.
    For user `https://live.fc2.com/24374512/`, user id would be `24374512`.

    Since the endpoint used for monitoring does not provide the user's nickname,
    the name of the configuration entity is used instead.
    """


    async def get_new_records(self, entity: FC2MonitorEntity, session: aiohttp.ClientSession) -> Sequence[FC2Record]:
        record = await self.check_channel(entity, session)
        return [record] if record else []

    async def check_channel(self, entity: FC2MonitorEntity, session: aiohttp.ClientSession) -> Optional[FC2Record]:
        data = await self.get_metadata(entity, session)
        if data is None:
            return None
        try:
            record = self.parse_metadata(data)
            if record is None:
                return None
        except (KeyError, TypeError, JSONDecodeError, pydantic.ValidationError) as e:
            self.logger.warning(f'FC2Monitor for {entity.name}: failed to parse channel info. Raw response: {data}')
            return None
        if record.start_timestamp == entity.latest_live_start:
            self.logger.debug(f'FC2Monitor for {entity.name}: user {entity.user_id} is live since {entity.latest_live_start}, but record was already created')
            return None
        self.logger.debug(f'FC2Monitor for {entity.name}: user {entity.user_id} is live since {record.start_timestamp}, producing record')
        entity.latest_live_start = record.start_timestamp
        record.name = entity.name
        return record

    async def get_metadata(self, entity: FC2MonitorEntity, session: aiohttp.ClientSession) -> Optional[str]:
        url = 'https://live.fc2.com/api/memberApi.php'
        data = {'channel': 1, 'streamid': entity.user_id}
        text = await self.request(url, entity, session, method='POST', data=data)
        return text

    @staticmethod
    def parse_metadata(raw_data: str) -> Optional[FC2Record]:
        data = json.loads(raw_data)
        data = data['data']['channel_data']
        is_live = data['is_publish']
        if not is_live:
            return None
        start_timestamp = str(data['start'])
        start = parse_timestamp_ms(start_timestamp)
        title = data['title']
        info = data['info']
        avatar_url = data['image']
        login_only = data['login_only']

        channel_id = str(data['channelid'])
        channel_url = f'https://live.fc2.com/{channel_id}/'

        return FC2Record(url=channel_url,
                         title=title,
                         user_id=channel_id,
                         start_timestamp=start_timestamp,
                         start=start,
                         info=info,
                         avatar_url=avatar_url,
                         login_only=login_only)



import abc
import asyncio
import datetime
import hashlib
import http.cookies
import json
import logging
import os
import re
import time
from collections import OrderedDict
from contextlib import ContextDecorator
from email.utils import mktime_tz, parsedate_to_datetime
from enum import Enum
from http import cookiejar
from math import log2
from pathlib import Path
from textwrap import shorten
from time import perf_counter
from typing import Any, Callable, Dict, Hashable, List, Optional, Tuple, Union

import aiohttp
import lxml.html
import multidict
from jsonpath import JSONPath
from multidict import CIMultiDictProxy

from avtdl.core.interfaces import Record


def load_cookies(path: Optional[Path], raise_on_error: bool = False) -> Optional[cookiejar.CookieJar]:
    """load cookies from a text file in Netscape format"""
    logger = logging.getLogger('cookies')
    if path is None:
        return None
    cookie_jar = cookiejar.MozillaCookieJar(path)
    try:
        cookie_jar.load(ignore_discard=True, ignore_expires=True)
        logger.info(f"Successfully loaded cookies from {path}")
    except FileNotFoundError:
        logger.exception(f'Failed to load cookies from {path}: file not found')
        if raise_on_error:
            raise
        return None
    except (cookiejar.LoadError, OSError) as e:
        if raise_on_error:
            raise
        logger.exception(f'Failed to load cookies from {path}: {e}')
        return None
    return cookie_jar


def convert_cookiejar(cookie_jar: cookiejar.CookieJar) -> aiohttp.CookieJar:
    """convert cookie jar produced by stdlib to format used by aiohttp"""
    cookies: http.cookies.BaseCookie = http.cookies.BaseCookie()
    for cookie in cookie_jar:
        name = cookie.name
        cookies[name] = cookie.value or ''
        cookies[name]['domain'] = cookie.domain
        cookies[name]['path'] = cookie.path
        cookies[name]['expires'] = str(cookie.expires)
        cookies[name]['secure'] = cookie.secure
        cookies[name]['version'] = str(cookie.version)
        cookies[name]['comment'] = cookie.comment
    new_jar = aiohttp.CookieJar(quote_cookie=False)
    new_jar.update_cookies(cookies)
    return new_jar


def get_cache_ttl(headers: multidict.CIMultiDictProxy) -> Optional[int]:
    """
    check for Expires and Cache-Control headers, return integer representing
    how many seconds is left until resource is outdated, if they are present
    and was parsed successfully
    """

    def get_expires_from_cache_control(headers: multidict.CIMultiDictProxy) -> Optional[datetime.datetime]:
        try:
            cache_control = headers.get('Cache-Control', '')
            [max_age] = re.findall('max-age=(\d+)', cache_control, re.IGNORECASE)
            max_age_value = datetime.timedelta(seconds=int(max_age))

            last_modified = headers.get('Last-Modified', '')
            last_modified_value = parsedate_to_datetime(last_modified)

            expires = last_modified_value + max_age_value
            return expires
        except (TypeError, ValueError, IndexError):
            return None

    def get_expires_from_expires_header(headers: multidict.CIMultiDictProxy) -> Optional[datetime.datetime]:
        try:
            expires_header = headers.get('Expires')
            if expires_header is None or expires_header == '0':
                return None
            expires_value = parsedate_to_datetime(expires_header)
            return expires_value
        except (TypeError, ValueError):
            return None

    expires = get_expires_from_cache_control(headers) or get_expires_from_expires_header(headers)
    if expires is None:
        return None

    now = datetime.datetime.now(tz=datetime.timezone.utc)
    delta = (expires - now).total_seconds()
    if delta < 0:
        return None

    return int(delta)


def get_retry_after(headers: Union[Dict[str, str], multidict.CIMultiDictProxy[str]]) -> Optional[int]:
    """return parsed value of Retry-After header, if present"""
    retry_after = headers.get('Retry-After')
    if retry_after is None:
        return None
    try:
        return int(retry_after)
    except ValueError:
        pass
    try:
        retry_at = parsedate_to_datetime(retry_after)
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        delay = int((retry_at - now).total_seconds())
        if delay > 0:
            return delay
    except (TypeError, ValueError):
        pass
    return None


def check_dir(path: Path, create=True) -> bool:
    """check if directory exists and writable, create if asked"""
    if path.is_dir() and os.access(path, mode=os.W_OK):
        return True
    elif create:
        logging.info(f'directory {path} does not exist, creating')
        try:
            path.mkdir(parents=True, exist_ok=True)
            return True
        except OSError as e:
            logging.warning(f'failed to create directory at {path}: {e}')
            return False
    else:
        return False


def make_datetime(items) -> datetime.datetime:
    """take 10-tuple and return datetime object with UTC timezone"""
    if len(items) == 9:
        items = *items, None
    if len(items) != 10:
        raise ValueError(f'Expected tuple with 10 elements, got {len(items)}')
    timestamp = mktime_tz(items)
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)


def parse_timestamp_us(timestamp: Union[str, int, None], ) -> Optional[datetime.datetime]:
    return parse_timestamp(timestamp, 6)


def parse_timestamp_ms(timestamp: Union[str, int, None], ) -> Optional[datetime.datetime]:
    return parse_timestamp(timestamp, 3)


def parse_timestamp(timestamp: Union[str, int, None], fraction: int) -> Optional[datetime.datetime]:
    """parse UNIX timestamp as datetime.datetime"""
    if timestamp is None:
        return None
    try:
        ts = int(timestamp)
        dt = datetime.datetime.fromtimestamp(int(ts / 10 ** fraction), tz=datetime.timezone.utc)
        return dt
    except Exception:
        return None


def show_diff(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> str:
    """pretty-print keys which values in dict1 and dict2 are different"""
    keys = {*dict1.keys(), *dict2.keys()}
    diff = []
    for k in keys:
        v1 = str(dict1.get(k, ''))
        repr_v1 = shorten(v1, 60)
        v2 = str(dict2.get(k, ''))
        repr_v2 = shorten(v2, 60)
        if v1 != v2:
            diff.append(f'[{k[:12]:12}]: {repr_v2:60} |->| {repr_v1:60}')
    return '\n'.join(diff)


async def request_raw(url: str, session: Optional[aiohttp.ClientSession], logger: Optional[logging.Logger] = None,
                      method: str = 'GET', params: Optional[Any] = None, data: Optional[Any] = None,
                      headers: Optional[Dict[str, Any]] = None, retry_times: int = 1, retry_delay: float = 1,
                      retry_multiplier: int = 1,
                      raise_errors: bool = False, raise_for_status: bool = True) -> Optional[aiohttp.ClientResponse]:
    logger = logger if logger else logging.getLogger('request')
    current_retry_delay = retry_delay
    for attempt in range(0, retry_times + 1):
        last_attempt = attempt == retry_times
        try:
            request: Callable = aiohttp.request
            if session is not None:
                request = session.request
                if session.headers is not None and headers is not None:
                    headers.update(session.headers)

            async with request(method=method, url=url, headers=headers, params=params, data=data) as response:
                if raise_for_status:
                    response.raise_for_status()
                _ = await response.text()
                return response

        except Exception as e:
            if not last_attempt:
                logger.warning(f'error when requesting {url}: {e}, retrying in {current_retry_delay:.02f} seconds')
                await asyncio.sleep(current_retry_delay)
                current_retry_delay *= retry_multiplier
                continue
            elif raise_errors:
                raise
            else:
                logger.warning(f'error when requesting {url}: {e}')
                return None
    return None


async def request(url: str, session: Optional[aiohttp.ClientSession] = None, logger: Optional[logging.Logger] = None,
                  method: str = 'GET', params: Optional[Any] = None, data: Optional[Any] = None,
                  headers: Optional[Dict[str, Any]] = None, retry_times: int = 1, retry_delay: float = 1,
                  retry_multiplier: int = 1,
                  raise_errors: bool = False) -> Optional[str]:
    logger = logger if logger else logging.getLogger('request')
    response = await request_raw(url, session, logger, method, params, data, headers, retry_times, retry_delay, retry_multiplier, raise_errors)
    if response is None:
        return None
    if response.status >= 300:
        # presuming that there is no point in retrying 3xx and aiohttp will follow redirects transparently
        logger.debug(f'got {response.status} ({response.reason}) from {url}')
        return None
    return await response.text()


async def request_json(url: str, session: Optional[aiohttp.ClientSession], logger: Optional[logging.Logger] = None,
                       method: str = 'GET', params: Optional[Any] = None, data: Optional[Any] = None,
                       headers: Optional[Dict[str, Any]] = None, retry_times: int = 1, retry_delay: float = 1,
                       retry_multiplier: int = 1,
                       raise_errors: bool = False) -> Optional[Any]:
    logger = logger if logger else logging.getLogger('request_json')
    text = await request(url, session, logger, method, params, data, headers, retry_times, retry_delay, retry_multiplier, raise_errors)
    if text is None:
        return None
    try:
        parsed = json.loads(text)
        return parsed
    except json.JSONDecodeError as e:
        logger.debug(f'error parsing response from {url}: {e}. Raw response data: "{text}"')
        if raise_errors:
            raise
        else:
            return None


class Delay:
    """Provide method to calculate next delay for exponential backoff based on S-shaped curve"""

    A: float = 4600  # upper asymptote
    k: float = 0.88  # curve growth rate
    x0: float = 8  # x value corresponding to midpoint of the curve

    @classmethod
    def _sigmoid(cls, x: float) -> float:
        y = cls.A / (1 + 2 ** (-cls.k * (x - cls.x0)))
        return y

    @classmethod
    def _inv_sigmoid(cls, y: float) -> float:
        if y <= 0:
            return 0
        # raises ValueError if y >= cls.A
        x = cls.x0 - log2((cls.A - y) / y) / cls.k
        return x

    @classmethod
    def get_next(cls, current: float) -> float:
        """
        Find current value on S-shaped curve and return a next one

        Returns 0 if input is not higher than 0,
        returns cls.A if input higher than cls.A.
        """
        try:
            current_step = cls._inv_sigmoid(current)
        except ValueError:
            current_step = current
        next_step = current_step + 1
        next_delay = cls._sigmoid(next_step)
        return next_delay


class timeit(ContextDecorator):
    """measure time call takes, print it in the log"""

    def __init__(self, logger: Optional[logging.Logger] = None):
        self.start: float = 0
        self.end: float = 0
        self.logger = logger

    @property
    def duration(self) -> float:
        return self.end - self.start

    @property
    def timedelta(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self.duration)

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = perf_counter()
        if self.logger is not None:
            self.logger.debug(f'took {self.timedelta}')
        return False


class LRUCache:

    def __init__(self, max_size: int = 100):
        if max_size <= 0:
            raise ValueError('Maximum cache size must be a positive integer')
        self._max_size = max_size
        self._data: OrderedDict = OrderedDict()

    def put(self, item: Hashable):
        """Put item in the cache, resize the cache if needed"""
        if not item in self._data:
            self._data[item] = 1
        self._data.move_to_end(item)
        while len(self._data) > self._max_size:
            self._data.popitem(last=False)


def find_matching_field(record: Record, pattern: str, fields: Optional[List[str]] = None) -> Optional[str]:
    name, _ = find_matching_field_name_and_value(record, pattern, fields)
    return name


def find_matching_field_value(record: Record, pattern: str, fields: Optional[List[str]] = None) -> Optional[str]:
    _, value = find_matching_field_name_and_value(record, pattern, fields)
    return value


def find_matching_field_name_and_value(record: Record, pattern: str, fields: Optional[List[str]] = None) -> Tuple[Optional[str], Optional[Any]]:
    """
    Return name of the first field of the record that contains pattern,
    return None if nothing found. If fields value specified only check
    fields listed in there.
    """
    for field, value in record:
        if fields is not None and field not in fields:
            continue
        if isinstance(value, Record):
            subrecord_search_result = find_matching_field_name_and_value(value, pattern)
            if subrecord_search_result is not None:
                return subrecord_search_result
        else:
            if str(value).find(pattern) > -1:
                return field, value
    return None, None


def record_has_text(record: Record, text: str) -> bool:
    return find_matching_field(record, text) is not None


def sanitize_filename(name: str) -> str:
    """Replace symbols not allowed in file names on NTFS with underscores"""
    return re.sub(r'[\\/:*?"<>|]', "_", name)


class OutputFormat(str, Enum):
    str = 'text'
    repr = 'short'
    json = 'json'
    pretty_json = 'pretty_json'
    hash = 'hash'


class Fmt:
    """Helper class to interpolate format string from config using data from Record"""

    @classmethod
    def format(cls, fmt: str, record: Record, missing: Optional[str] = None, tz: Optional[datetime.timezone] = None, sanitize: bool = False, extra: Optional[Dict[str, Any]] = None) -> str:
        """Take string with placeholders like {field} and replace them with record fields"""
        logger = logging.getLogger().getChild('format')
        result = cls.strftime(fmt, datetime.datetime.now(tz))
        record_as_dict = record.model_dump()
        if extra is not None:
            record_as_dict.update(extra)
        placeholders: List[str] = re.findall(r'({[^{}\\]+})', fmt)
        for placeholder in placeholders:
            field = placeholder.strip('{}')
            value = record_as_dict.get(field)
            if value is not None:
                value = cls.format_value(value, sanitize)
                result = result.replace(placeholder, value)
            else:
                if missing is not None:
                    result = result.replace(placeholder, missing)
                else:
                    logger.warning(f'placeholder "{placeholder}" used by format string "{fmt}" is not a field of {record.__class__.__name__} ({record!r}), resulting command is unlikely to be valid')
        result = result.replace(r'\{', '{')
        result = result.replace(r'\}', '}')
        return result
    
    @classmethod
    def format_value(cls, value: Any, sanitize: bool = False) -> str:
        if value is None:
            value = ''
        elif isinstance(value, datetime.datetime):
            value = cls.date(value)
        else:
            value = str(value)
            if sanitize:
                value = sanitize_filename(value)
        return value

    @classmethod
    def format_path(cls, path: Union[str, Path], record: Record, missing: Optional[str] = None, tz: Optional[datetime.timezone] = None, extra: Optional[Dict[str, Any]] = None) -> Path:
        """Take string with placeholders and replace them with record fields, but strip them from bad symbols"""
        fmt = str(path)
        formatted_path = cls.format(fmt, record, missing, tz=tz, sanitize=True, extra=extra)
        return Path(formatted_path)

    @classmethod
    def strftime(cls, fmt: str, dt: datetime.datetime) -> str:
        if '%' in fmt:
            fmt = re.sub(r'(%[^aAwdbBmyYHIpMSfzZjUWcxX%GuV])', r'%\1', fmt)
        try:
            return dt.strftime(fmt)
        except ValueError as e:
            logger = logging.getLogger().getChild('format').getChild('strftime')
            logger.debug(f'error adding current date to template "{fmt}": {e}')
            return fmt

    @classmethod
    def date(cls, dt: datetime.datetime) -> str:
        return dt.strftime('%Y-%m-%d %H:%M')

    @classmethod
    def dtf(cls, dt: datetime.datetime) -> str:
        """format datetime to Discord timestamp"""
        ts = int(dt.timestamp())
        return f'<t:{ts}>'

    @classmethod
    def save_as(cls, record: Record, output_format: OutputFormat = OutputFormat.str) -> str:
        """Take a record and convert in to string as text/json or sha1"""
        if output_format == OutputFormat.str:
            return str(record)
        if output_format == OutputFormat.repr:
            return repr(record)
        if output_format == OutputFormat.json:
            return record.as_json()
        if output_format == OutputFormat.pretty_json:
            return record.as_json(2)
        if output_format == OutputFormat.hash:
            return record.hash()


def html_to_text(html: str) -> str:
    """take html fragment, try to parse it and extract text values using lxml"""
    try:
        root = lxml.html.fromstring(html)

        # text_content() skips <img> content altogether
        # walk tree manually and for images containing links
        # add them to text representation
        for elem in root.iter():
            if elem.tag == 'img':
                image_link = elem.get('src')
                if image_link is not None:
                    elem.text = f'\n{image_link}\n'
        text = root.text_content()
        return text
    except Exception as e:
        logger = logging.getLogger('html_to_text')
        logger.warning(e)
        return html


def read_file(path: Union[str, Path], encoding=None) -> str:
    """
    Read and return file content in provided encoding

    If decoding file content in provided encoding fails, try again using utf8.
    If it also fails, let the exception to propagate. Handling OSError is also
    left to caller.
    """
    with open(path, 'rt', encoding=encoding) as fp:
        try:
            text = fp.read()
            return text
        except UnicodeDecodeError:
            pass
    with open(path, encoding='utf8') as fp:
        text = fp.read()
        return text


def write_file(path: Union[str, Path], content: str, encoding='utf8', backups: int = 0):
    if backups > 0:
        rotate_file(path, depth=backups)
    with open(path, 'wt', encoding=encoding) as fp:
        fp.write(content)


def rotate_file(path: Union[str, Path], depth: int = 10):
    """Move "path" to "path.1", "path.1" to "path.2" and so on down to depth parameter"""
    increment_postfix(path, depth)


def increment_postfix(path: Union[str, Path], maxdepth):
    path = Path(path)
    if not path.exists():
        return
    if re.match('\.(\d|[1-9]\d+)$', path.suffix):
        index = int(path.suffix.strip('.'))
        next_path = path.with_suffix(f'.{index + 1}')
    else:
        index = 0
        next_path = path.with_suffix(path.suffix + '.0')
    if index >= maxdepth:
        return
    increment_postfix(next_path, maxdepth)
    logging.getLogger('rotate').info(f'moving {path} to {next_path}')
    path.replace(next_path)


def sha1(text: str) -> str:
    return hashlib.sha1(text.encode()).digest().hex()


def get_cookie_value(jar: aiohttp.CookieJar, key: str) -> Optional[str]:
    for morsel in jar:
        if morsel.key == key:
            return morsel.value
    return None


def find_all(data: Union[dict, list], jsonpath: str, cache={}) -> list:
    if jsonpath not in cache:
        cache[jsonpath] = JSONPath(jsonpath)
    parser = cache[jsonpath]
    return parser.parse(data)


def find_one(data: Union[dict, list], jsonpath: str) -> Optional[Any]:
    result = find_all(data, jsonpath)
    return result[0] if result else None


class SessionStorage:
    """
    Provide way to initialize, store and reuse ClientSession objects

    The "name" parameter can be used to get distinct sessions
    with the same cookies and headers.

    To ensure sessions shared between multiple tasks are safely closed,
    user must call SessionStorage.run() once from a running event loop
    and should not use "async with" on any of the session objects, as it
    might cause the session to be closed prematurely, making other tasks
    using it fail.
    """

    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.sessions: Dict[str, aiohttp.ClientSession] = {}
        self.task: Optional[asyncio.Task] = None
        self.logger = (logger or logging.getLogger()).getChild('sessions')

    @staticmethod
    def get_session_id(cookies_file: Optional[Path], headers: Optional[Dict[str, Any]], name: str = '') -> str:
       return name +  str((cookies_file, headers))

    def get_session_by_id(self, session_id: str) -> Optional[aiohttp.ClientSession]:
        return self.sessions.get(session_id)

    def session_exists(self, cookies_file: Optional[Path], headers: Optional[Dict[str, Any]], name: str = '') -> bool:
        session_id = self.get_session_id(cookies_file, headers, name)
        session = self.get_session_by_id(session_id)
        return session is not None

    def get_session(self, cookies_file: Optional[Path] = None, headers: Optional[Dict[str, Any]] = None, name: str = '') -> aiohttp.ClientSession:
        session_id = self.get_session_id(cookies_file, headers, name)
        session = self.get_session_by_id(session_id)
        if session is None:
            netscape_cookies = load_cookies(cookies_file)
            cookies = convert_cookiejar(netscape_cookies) if netscape_cookies else None
            session = aiohttp.ClientSession(cookie_jar=cookies, headers=headers)
            self.sessions[session_id] = session
        return session

    async def ensure_closed(self) -> None:
        try:
            await asyncio.Future()
        except (asyncio.CancelledError, KeyboardInterrupt):
            self.logger.debug('closing http sessions...')
            for session_id, session in self.sessions.items():
                if not session.closed:
                    self.logger.debug(f'closing session "{session_id}"')
                    await session.close()
            self.logger.debug('done')

    def run(self) -> None:
       if self.task is None:
           name = f'ensure_closed for {self.logger.name} ({self!r})'
           self.task = asyncio.create_task(self.ensure_closed(), name=name)


class RateLimit:
    """
    Encapsulate state of rate limits for "bucket of tokens" type of endpoint

    >>> async def endpoint_request(url):
    >>>     async with RateLimit('endpoint name') as rate_limit:
    >>>         response = await request_raw(url)
    >>>         rate_limit.submit_headers(response.headers)
    >>>

    Request itself, with error handling, is done by the endpoint class,
    which then should then call RateLimit.submit_headers with response headers
    to update the limit values and expiration time.
    """

    def __init__(self, name: str, logger: Optional[logging.Logger] = None) -> None:
        self.limit_total: int = 50
        self.limit_remaining: int = 10
        self.reset_at: int = int(datetime.datetime.now(tz=datetime.timezone.utc).timestamp())

        self.name = name
        self.logger = logger or logging.getLogger().getChild('rate_limit')
        self.lock = asyncio.Lock()
        self.perf_lock_acquired_at: float = 0

    async def __aenter__(self) -> 'RateLimit':
        with timeit() as t:
            await self.lock.acquire()
        self.perf_lock_acquired_at = t.end
        if self.limit_remaining <= 1:
            self.logger.debug(f'[{self.name}] lock acquired in {t.timedelta}, {self.delay} seconds until rate limit reset')
            await asyncio.sleep(self.delay)
        else:
            self.logger.debug(f'[{self.name}] lock acquired in {t.timedelta}, there is no active rate limit')
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()
        duration = datetime.timedelta(seconds=(time.perf_counter() - self.perf_lock_acquired_at))
        self.logger.debug(f'[{self.name}] lock released after {duration}')
        return False

    @property
    def reset_at_date(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self.reset_at, tz=datetime.timezone.utc)

    @property
    def reset_after(self) -> int:
        now = int(datetime.datetime.now().timestamp())
        reset_after = max(0, self.reset_at - now)
        return reset_after

    @property
    def delay(self) -> int:
        if self.limit_remaining <= 1:
            return self.reset_after + 1
        return 0

    @abc.abstractmethod
    def _submit_headers(self, headers: Union[Dict[str, str], CIMultiDictProxy[str]], logger: logging.Logger):
        """parse response headers and update self.limit_total, self.limit_remaining and self.reset_at"""

    def submit_headers(self, headers: Union[Dict[str, str], CIMultiDictProxy[str]], logger: Optional[logging.Logger] = None):
        """
        Update limits values and reset time using data from headers.

        Client code should call it after completing request. Client code might
        provide a custom Logger instance to help with identifying debug output
        from specific request.
        """
        logger = logger or self.logger

        retry_after = get_retry_after(headers)
        if retry_after is not None:
            self.limit_remaining = 0
            self.reset_at = int((datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(seconds=retry_after)).timestamp())
            logger.debug(f'[{self.name}] Retry-After is present and set to {retry_after}, setting reset_at to {self.reset_at}')
            return

        self._submit_headers(headers, logger)


def strip_text(s: str, text: str) -> str:
    if s.startswith(text):
        return s[len(text):]
    return s



import asyncio
import json
from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple

import aiohttp
from pydantic import Field, FilePath, PositiveFloat, field_serializer, field_validator

from avtdl.core.db import BaseDbConfig, RecordDB
from avtdl.core.interfaces import ActorConfig, Monitor, MonitorEntity, Record, RuntimeContext
from avtdl.core.utils import Delay, SessionStorage, get_cache_ttl, get_retry_after, load_cookies, \
    show_diff

HIGHEST_UPDATE_INTERVAL = 4 * 3600


class TaskMonitorEntity(MonitorEntity):
    update_interval: PositiveFloat
    """how often the monitored source should be checked for new content, in seconds"""


class BaseTaskMonitor(Monitor):

    def __init__(self, conf: ActorConfig, entities: Sequence[TaskMonitorEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

    async def run(self):
        await self.start_cyclic_tasks()

    async def start_cyclic_tasks(self):
        by_entity_interval = defaultdict(list)
        for entity in self.entities.values():
            by_entity_interval[entity.update_interval].append(entity)
        by_group_interval = defaultdict(list)
        for interval, entities in by_entity_interval.items():
            by_group_interval[interval / len(entities)].extend(entities)
        for interval in sorted(by_group_interval.keys()):
            entities = by_group_interval[interval]
            _ = self.controller.create_task(self.start_tasks_for(entities, interval), name=f'start_cyclic_tasks_{interval}')

    async def start_tasks_for(self, entities: List[TaskMonitorEntity], interval: float) -> None:
        assert self.logger.parent is not None
        logger = self.logger.parent.getChild('scheduler').getChild(self.conf.name)
        if len(entities) == 0:
            logger.debug(f'called with no entities and {interval} interval')
            return
        names = ', '.join([f'{self.conf.name}.{entity.name}' for entity in entities])
        logger.info(f'will start {len(entities)} tasks with {entities[0].update_interval:.1f} update interval and {interval:.1f} offset for {names}')
        for entity in entities:
            logger.debug(f'starting task {entity.name} with {entity.update_interval} update interval')
            _ = self.controller.create_task(self.run_for(entity), name=f'{self.conf.name}:{entity.name}')
            if entity == entities[-1]: # skip sleeping after last
                continue
            await asyncio.sleep(interval)
        logger.info(f'done starting tasks for {names}')

    @abstractmethod
    async def run_for(self, entity: TaskMonitorEntity):
        '''Task for a specific entity that should check for new records based on update_interval and call self.on_record() for each'''


class TaskMonitor(BaseTaskMonitor):

    async def run_for(self, entity: TaskMonitorEntity):
        while True:
            try:
                await self.run_once(entity)
            except Exception:
                self.logger.exception(f'{self.conf.name}: task for entity {entity} failed, terminating')
                break
            await asyncio.sleep(entity.update_interval)

    async def run_once(self, entity: TaskMonitorEntity):
        records = await self.get_new_records(entity)
        for record in records:
            self.on_record(entity, record)

    @abstractmethod
    async def get_new_records(self, entity: TaskMonitorEntity) -> Sequence[Record]:
        '''Produce new records, optionally adjust update_interval'''


class HttpTaskMonitorEntity(TaskMonitorEntity):
    cookies_file: Optional[FilePath] = None
    """path to a text file containing cookies in Netscape format"""
    headers: Optional[Dict[str, str]] = {'Accept-Language': 'en-US,en;q=0.9'}
    """custom HTTP headers as pairs "key": value". "Set-Cookie" header will be ignored, use `cookies_file` option instead"""

    adjust_update_interval: bool = True
    """change delay before next update based on response headers. This setting doesn't affect timeouts after failed requests"""
    base_update_interval: float = Field(exclude=True, default=60)
    """internal variable to persist state between updates. Used to keep update_interval while timeout after update error is active"""
    last_modified: Optional[str] = Field(exclude=True, default=None)
    """internal variable to persist state between updates. Used to keep Last-Modified header value"""
    etag: Optional[str] = Field(exclude=True, default=None)
    """internal variable to persist state between updates. Used to keep Etag header value"""

    @field_validator('cookies_file')
    @classmethod
    def check_cookies(cls, path: Optional[Path]):
        if path is None:
            return None
        try:
            load_cookies(path, raise_on_error=True)
        except Exception as e:
            raise ValueError(f'{e}') from e
        return path

    @field_serializer('update_interval')
    def restore_update_interval(self, _: float) -> float:
        return self.base_update_interval

    def model_post_init(self, __context: Any) -> None:
        self.base_update_interval = self.update_interval


class HttpTaskMonitor(BaseTaskMonitor):
    '''Maintain and provide for records aiohttp.ClientSession objects
    grouped by HttpTaskMonitorEntity.cookies_path, which means entities that use
    the same cookies file will share session'''

    def __init__(self, conf: ActorConfig, entities: Sequence[HttpTaskMonitorEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.sessions: SessionStorage = SessionStorage(self.logger)

    async def request_json(self, url: str, entity: HttpTaskMonitorEntity, session: aiohttp.ClientSession, method='GET', headers: Optional[Dict[str, str]] = None, params: Optional[Mapping] = None, data: Optional[Any] = None, data_json: Optional[Any] = None) -> Optional[Any]:
        text = await self.request(url, entity, session, method, headers, params, data, data_json)
        if text is None:
            return None
        try:
            parsed = json.loads(text)
            return parsed
        except json.JSONDecodeError as e:
            self.logger.debug(f'error parsing response from {url}: {e}. Raw response data: "{text}"')
            return None

    async def request(self, url: str, entity: HttpTaskMonitorEntity, session: aiohttp.ClientSession, method='GET', headers: Optional[Dict[str, str]] = None, params: Optional[Mapping] = None, data: Optional[Any] = None, data_json: Optional[Any] = None) -> Optional[str]:
        response = await self.request_raw(url, entity, session, method, headers, params, data, data_json)
        if response is None:
            return None
        return await response.text()

    async def request_raw(self, url: str, entity: HttpTaskMonitorEntity, session: aiohttp.ClientSession, method='GET', headers: Optional[Dict[str, str]] = None, params: Optional[Mapping] = None, data: Optional[Any] = None, data_json: Optional[Any] = None) -> Optional[aiohttp.ClientResponse]:
        '''Helper method to make http request. Does not retry, adjusts entity.update_interval instead'''
        if self.logger.parent is None:
            # should never happen since Actor().logger is constructed with getChild()
            logger = self.logger.getChild('request').getChild(self.conf.name)
        else:
            logger = self.logger.parent.getChild('request').getChild(self.conf.name)
        request_headers: Dict[str, Any] = headers or {}
        if session.headers is not None:
            request_headers.update(session.headers)
        if entity.last_modified is not None and method in ['GET', 'HEAD']:
            request_headers['If-Modified-Since'] = entity.last_modified
        if entity.etag is not None:
            request_headers['If-None-Match'] = entity.etag
        try:
            text = ''
            async with session.request(method, url, headers=request_headers, params=params, data=data, json=data_json) as response:
                # fully read http response to get it cached inside ClientResponse object
                # client code can then use it by awaiting .text() again without causing
                # network activity and potentially triggering associated errors
                text = await response.text()
                response.raise_for_status()
        except Exception as e:
            if isinstance(e, aiohttp.ClientResponseError):
                logger.warning(f'[{entity.name}] got code {e.status} ({e.message}) while fetching {url}')
                if text:
                    logger.debug(f'[{entity.name}] response body: "{text}"')
                retry_after = get_retry_after(response.headers)
                if retry_after is not None:
                    raw_header = response.headers.get("Retry-After")
                    logger.debug(f'[{entity.name}] got Retry-After header with value {raw_header}')
                    entity.update_interval = max(float(retry_after), HIGHEST_UPDATE_INTERVAL)
                    logger.warning(f'[{entity.name}] update interval set to {entity.update_interval} seconds for {url} as requested by response headers')
                    return None
            else:
                logger.warning(f'[{entity.name}] error while fetching {url}: {e.__class__.__name__} {e}')

            update_interval = int(max(Delay.get_next(entity.update_interval), entity.update_interval))
            if entity.update_interval != update_interval:
                entity.update_interval = update_interval
                logger.warning(f'[{entity.name}] update interval set to {entity.update_interval} seconds for {url}')
            return None

        if response.status == 304:
            logger.debug(f'[{entity.name}] got {response.status} ({response.reason}) from {url}')
            if entity.update_interval != entity.base_update_interval:
                logger.info(f'[{entity.name}] restoring update interval {entity.base_update_interval} seconds for {url} after getting 304 response')
                entity.update_interval = entity.base_update_interval
            return None
        # some servers do not have cache headers in 304 response, so only updating on 200
        entity.last_modified = response.headers.get('Last-Modified', None)
        entity.etag = response.headers.get('Etag', None)

        cache_control = response.headers.get('Cache-control')
        logger.debug(f'[{entity.name}] Last-Modified={entity.last_modified or "absent"}, ETAG={entity.etag or "absent"}, Cache-control="{cache_control or "absent"}"')

        if entity.adjust_update_interval:
            new_update_interval = get_cache_ttl(response.headers) or entity.base_update_interval
            new_update_interval = min(new_update_interval, 10 * entity.base_update_interval, HIGHEST_UPDATE_INTERVAL) # in case ttl is overly long
            new_update_interval = max(new_update_interval, entity.base_update_interval)
            if entity.update_interval != new_update_interval:
                entity.update_interval = new_update_interval
                logger.info(f'[{entity.name}] next update in {entity.update_interval}')
        else:
            # restore update interval after backoff on failure
            if entity.update_interval != entity.base_update_interval:
                logger.info(f'[{entity.name}] restoring update interval {entity.base_update_interval} seconds for {url}')
                entity.update_interval = entity.base_update_interval

        return response

    def _get_session(self, entity: HttpTaskMonitorEntity) -> aiohttp.ClientSession:
        session_id = self.sessions.get_session_id(entity.cookies_file, entity.headers)
        session = self.sessions.get_session_by_id(session_id)
        if session is None:
            session = self.sessions.get_session(entity.cookies_file, entity.headers)
        else:
            self.logger.debug(f'[{entity.name}] reusing session with cookies from {session_id}')
        return session

    async def run(self):
        name = f'ensure_closed for {self.logger.name} ({self!r})'
        _ = self.controller.create_task(self.sessions.ensure_closed(), name=name)
        await super().run()

    async def run_for(self, entity: HttpTaskMonitorEntity):
        try:
            session = self._get_session(entity)
            while True:
                await self.run_once(entity, session)
                await asyncio.sleep(entity.update_interval)
        except Exception:
            self.logger.exception(f'unexpected error in task for entity {entity.name}, task terminated')

    async def run_once(self, entity: TaskMonitorEntity, session: aiohttp.ClientSession):
        records = await self.get_new_records(entity, session)
        for record in records:
            self.on_record(entity, record)

    @abstractmethod
    async def get_new_records(self, entity: TaskMonitorEntity, session: aiohttp.ClientSession) -> Sequence[Record]:
        '''Produce new records, optionally adjust update_interval'''


class BaseFeedMonitorConfig(BaseDbConfig):
    pass


class BaseFeedMonitorEntity(HttpTaskMonitorEntity):
    url: str
    """url that should be monitored"""

    quiet_start: bool = False
    """throw away new records on the first update after application startup"""
    quiet_first_time: bool = True
    """throw away new records produced on first update of given url"""

class BaseFeedMonitor(HttpTaskMonitor):

    def __init__(self, conf: BaseFeedMonitorConfig, entities: Sequence[BaseFeedMonitorEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.db = RecordDB(conf.db_path, logger=self.logger.getChild('db'))

    @abstractmethod
    async def get_records(self, entity: BaseFeedMonitorEntity, session: aiohttp.ClientSession) -> Sequence[Record]:
        '''Fetch and parse resource, return parsed records, both old and new'''

    async def run(self):
        for entity in self.entities.values():
            session = self._get_session(entity)
            await self.prime_db(entity, session)
        await super().run()

    async def prime_db(self, entity: BaseFeedMonitorEntity, session: aiohttp.ClientSession) -> None:
        '''if a feed has no prior records, fetch it once and mark all entries as old
        in order to not produce ten messages at once when the feed is first added'''
        size = self.db.get_size(entity.name)
        priming_required = False
        if entity.quiet_start:
            self.logger.info(f'[{entity.name}] option "quiet_start" enabled, all records until this moment will be marked as already seen')
            priming_required = True
        elif size == 0:
            self.logger.info(f'[{entity.name}] database at "{self.conf.db_path}" has no records for "{entity.name}", assuming first run')
            if entity.quiet_first_time:
                self.logger.debug(f'[{entity.name}] option "quiet_first_time" enabled, all records until this moment will be marked as already seen')
                priming_required = True
        if priming_required:
            n = len(await self.get_new_records(entity, session))
            self.logger.debug(f'[{entity.name}] number of records that was marked as already seen on first update: {n}')
        else:
            self.logger.info(f'[{entity.name}] {size} records stored in database')

    def store_records(self, records: Sequence[Record], entity: BaseFeedMonitorEntity):
        self.db.store_records(records, entity.name)

    def load_record(self, record: Record, entity: BaseFeedMonitorEntity) -> Optional[Record]:
        return self.db.load_record(record, entity.name)

    def record_is_new(self, record: Record, entity: BaseFeedMonitorEntity) -> bool:
        return not self.db.record_exists(record, entity.name)

    def record_got_updated(self, record: Record, entity: BaseFeedMonitorEntity) -> bool:
        return self.db.record_got_updated(record, entity.name)

    def _log_changes(self, record: Record, entity: BaseFeedMonitorEntity):
        normalized_record = type(record).model_validate_json(record.as_json())
        stored_record = self.load_record(record, entity)
        if stored_record is None:
            return
        stored_record_instance = type(record).model_validate_json(stored_record.as_json())
        msg = f'[{entity.name}] fetched record "{record.get_uid()}" (new: {record.hash()[:5]}, old: {stored_record_instance.hash()[:5]}) already exists but has changed:\n'
        self.logger.debug(msg + show_diff(normalized_record.model_dump(), stored_record_instance.model_dump()))

    def filter_new_records(self, records: Sequence[Record], entity: BaseFeedMonitorEntity) -> Sequence[Record]:
        new_records = []
        records_to_store = []
        for record in records:
            if self.record_is_new(record, entity):
                new_records.append(record)
                records_to_store.append(record)
                self.logger.debug(f'[{entity.name}] fetched record is new: "{record.get_uid()}" (hash: {record.hash()[:5]})')
            elif self.record_got_updated(record, entity):
                records_to_store.append(record)
                self._log_changes(record, entity)
                self.logger.debug(f'[{entity.name}] storing new version of record "{record.get_uid()}" (hash: {record.hash()[:5]})')
        self.store_records(records_to_store, entity)
        return new_records

    async def get_new_records(self, entity: BaseFeedMonitorEntity, session: aiohttp.ClientSession) -> Sequence[Record]:
        records = await self.get_records(entity, session)
        new_records = self.filter_new_records(records, entity)
        return new_records


class PagedFeedMonitorConfig(BaseFeedMonitorConfig):
    pass


class PagedFeedMonitorEntity(BaseFeedMonitorEntity):
    max_continuation_depth: int = 10
    """when updating feed with pagination support, only continue for this many pages"""
    next_page_delay: float = 1
    """when updating feed with pagination support, wait this much before loading next page"""
    allow_discontinuity: bool = False # store already fetched records on failure to load one of older pages
    """when updating feed with pagination support, if this setting is enabled and error happens when loading a page, records from already parsed pages will not be dropped. It will allow update of the feed to finish, but older records from deeper pages will then never be parsed on consecutive updates"""
    fetch_until_the_end_of_feed_mode: bool = False
    """when updating feed with pagination support, enables special mode, which makes a monitor try loading and parsing all pages until the end, even if they have been already parsed. Designed for purpose of archiving entire feed content"""

    def model_post_init(self, __context: Any) -> None:
        if self.fetch_until_the_end_of_feed_mode:
            self.quiet_first_time = False
            self.quiet_start = False
        super().model_post_init(__context)


class PagedFeedMonitor(BaseFeedMonitor, ABC):
    '''Provide support for loading and parsing feeds with pagination or lazy loading'''

    @abstractmethod
    async def handle_first_page(self, entity: PagedFeedMonitorEntity, session: aiohttp.ClientSession) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        '''Download and parse first page of the feed

        Returns two-elements tuple with processed records as first element and
        anything required to load and process next page as second.

        If loading or parsing page failed, warning is issued using self.logger, update_interval
        adjusted if required (by using self.request to fetch data or manually)
        and first element of the returned tuple is None.

        If there is no next page or there is no new records, or limit of continuation depth reached,
        then first element is an empty list and the second element is None'''

    @abstractmethod
    async def handle_next_page(self, entity: PagedFeedMonitorEntity, session: aiohttp.ClientSession, context: Optional[Any]) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        '''Download and parse continuation  page

        Parameters:
            entity (PagedFeedMonitorEntity): working entity
            session (aiohttp.ClientSession): session object to make requests with
            context (Optional[Any]): any data required to load next page, such as continuation token
        Returns same values as handle_first_page'''

    async def get_records(self, entity: PagedFeedMonitorEntity, session: aiohttp.ClientSession) -> Sequence[Record]:
        records: List[Record] = []
        current_page_records, continuation_context = await self.handle_first_page(entity, session)
        if current_page_records is None:
            return []
        records.extend(current_page_records)

        if entity.fetch_until_the_end_of_feed_mode:
            self.logger.info(
                f'[{entity.name}] "fetch_until_the_end_of_feed_mode" setting is enabled, will keep loading through already seen pages until the end. Disable it in config after it succeeds once')

        current_page = 1
        while True:
            if continuation_context is None:
                self.logger.debug(f'[{entity.name}] no continuation link on {current_page - 1} page, end of feed reached')
                if entity.fetch_until_the_end_of_feed_mode:
                    self.logger.info(f'[{entity.name}] reached the end of the feed at {current_page - 1} page, fetch_until_the_end_of_feed_mode can be disabled now')
                    entity.fetch_until_the_end_of_feed_mode = False
                break
            if not entity.fetch_until_the_end_of_feed_mode:
                if current_page > entity.max_continuation_depth:
                    self.logger.info(
                        f'[{entity.name}] reached continuation limit of {entity.max_continuation_depth}, aborting update')
                    break
                if not all(self.record_is_new(record, entity) for record in current_page_records):
                    self.logger.debug(f'[{entity.name}] found already stored records on {current_page - 1} page')
                    break
            self.logger.debug(f'[{entity.name}] all records on page {current_page - 1} are new, loading next one')

            current_page_records, continuation_context = await self.handle_next_page(entity, session, continuation_context)
            if current_page_records is None:
                if entity.allow_discontinuity or entity.fetch_until_the_end_of_feed_mode:
                    # when unable to load _all_ new records, return at least current progress
                    break
                else:
                    # when unable to load _all_ new records, throw away all already parsed and return nothing
                    # to not cause discontinuity in stored data
                    return []
            records.extend(current_page_records)
            self.logger.debug(f'[{entity.name}] while parsing page {current_page} got {len(current_page_records)} records')

            current_page += 1
            await asyncio.sleep(entity.next_page_delay)

        records = records[::-1]
        return records



import datetime
import logging
import sqlite3
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Set, Union

from pydantic import Field, field_validator, model_validator

from avtdl.core.interfaces import ActorConfig, Record
from avtdl.core.utils import check_dir


class BaseRecordDB:
    table_name = 'records'
    table_structure = 'parsed_at datetime, feed_name text, uid text, hashsum text, class_name text, as_json text, PRIMARY KEY(uid, hashsum)'
    row_structure = ':parsed_at, :feed_name, :uid, :hashsum, :class_name, :as_json'
    id_field = 'uid'
    exact_id_field = 'hashsum'
    group_id_field = 'feed_name'
    sorting_field = 'parsed_at'

    def __init__(self, db_path: Union[str, Path], logger: Optional[logging.Logger] = None):
        self.logger = logger or logging.getLogger('RecordDB')
        try:
            if not db_path == ':memory:' and not Path(db_path).exists():
                check_dir(Path(db_path).parent)
            self.db = sqlite3.connect(db_path)
            self.db.row_factory = sqlite3.Row
            self.cursor = self.db.cursor()
            self.cursor.execute('CREATE TABLE IF NOT EXISTS {} ({})'.format(self.table_name, self.table_structure))
            self.db.commit()
        except sqlite3.OperationalError as e:
            self.logger.error(
                f'error opening sqlite database at path "{db_path}", specified in "db_path" config variable: {e}. If file exists make sure it was produced by this application, otherwise check if new file can be created at specified location. Alternatively use special value ":memory:" to use in-memory database instead.')
            raise
        else:
            self.logger.debug(f'successfully connected to sqlite database at "{db_path}"')

    def store(self, rows: Union[Dict[str, Any], List[Dict[str, Any]]]) -> None:
        sql = "INSERT OR IGNORE INTO {} VALUES({})".format(self.table_name, self.row_structure)
        if not isinstance(rows, list):
            rows = [rows]
        self.cursor.executemany(sql, rows)
        self.db.commit()

    def fetch_row(self, uid: Any, exact_id: Optional[str] = None) -> Optional[sqlite3.Row]:
        if exact_id is not None:
            sql = f'SELECT * FROM records WHERE {self.id_field}=:uid AND {self.exact_id_field}=:exact_id ORDER BY {self.sorting_field} DESC LIMIT 1'
        else:
            sql = f'SELECT * FROM records WHERE {self.id_field}=:uid ORDER BY {self.sorting_field} DESC LIMIT 1'
        keys = {'uid': uid, 'exact_id': exact_id}
        self.cursor.execute(sql, keys)
        return self.cursor.fetchone()

    def row_exists(self, uid: Any, exact_id: Optional[str] = None) -> bool:
        if exact_id is not None:
            sql = f'SELECT 1 FROM records WHERE {self.id_field}=:uid AND {self.exact_id_field}=:exact_id LIMIT 1'
        else:
            sql = f'SELECT 1 FROM records WHERE {self.id_field}=:uid LIMIT 1'
        keys = {'uid': uid, 'exact_id': exact_id}
        self.cursor.execute(sql, keys)
        return self.cursor.fetchone() is not None

    def get_size(self, group_id: Optional[str] = None) -> int:
        '''return number of records, total or for specified feed, are stored in db'''
        if group_id is None:
            sql = f'SELECT COUNT(1) FROM {self.table_name}'
        else:
            sql = f'SELECT COUNT(1) FROM {self.table_name} WHERE {self.group_id_field}=:group'
        keys = {'group': group_id}
        self.cursor.execute(sql, keys)
        return int(self.cursor.fetchone()[0])


class RecordDB(BaseRecordDB):

    @staticmethod
    def _get_record_id(record: Record, entity_name: str) -> str:
        return '{}:{}'.format(entity_name, record.get_uid())

    def store_records(self, records: Sequence[Record], entity_name: str):
        rows = []
        for record in records:
            uid = self._get_record_id(record, entity_name)
            parsed_at = datetime.datetime.now(tz=datetime.timezone.utc)
            hashsum = record.hash()
            feed_name = entity_name
            class_name = record.__class__.__name__
            as_json = record.as_json()
            row = {'parsed_at': parsed_at, 'feed_name': feed_name, 'uid': uid, 'hashsum': hashsum, 'class_name': class_name, 'as_json': as_json}
            rows.append(row)
        self.store(rows)

    def load_record(self, record: Record, entity_name: str) -> Optional[Record]:
        """load most recently stored version of the record from db"""
        uid = self._get_record_id(record, entity_name)
        stored_record = self.fetch_row(uid)
        if stored_record is None:
            return None
        stored_record_instance = type(record).model_validate_json(stored_record['as_json'])
        return stored_record_instance

    def record_exists(self, record: Record, entity_name: str) -> bool:
        uid = self._get_record_id(record, entity_name)
        return self.row_exists(uid)

    def record_got_updated(self, record: Record, entity_name: str) -> bool:
        """return True when there are different versions of the record in db but not this one"""
        uid = self._get_record_id(record, entity_name)
        return self.row_exists(uid) and not self.row_exists(uid, record.hash())

    def record_has_changed(self, record: Record, entity_name: str, excluded_fields: Set[str]):
        """check if the record differs from most recently stored version, not counting fields listed in excluded_fields"""
        stored_record = self.load_record(record, entity_name)
        if stored_record is None:
            return False
        record_dump = record.model_dump(exclude=excluded_fields)
        stored_record_dump = stored_record.model_dump(exclude=excluded_fields)
        return record_dump != stored_record_dump


class BaseDbConfig(ActorConfig):
    db_path: Union[Path, str] = Field(default='db/', validate_default=True)
    """path to the sqlite database file keeping history of old records.
    Might specify a path to a directory containing the file (with trailing slash)
    or a direct path to the file itself (without a slash). If special value `:memory:` is used,
    database is kept in memory and not stored on disk at all, providing a clean database on every startup"""

    @field_validator('db_path')
    @classmethod
    def str_to_path(cls, path: Union[Path, str]):
        if isinstance(path, Path):
            return path
        if path == ':memory:':
            return path
        if path.endswith('/') or path.endswith('\\'):
            ok = check_dir(Path(path), create=True)
            if not ok:
                raise ValueError(f'error accessing path {path}, check if it is a valid path and is writeable')
        return Path(path)

    @model_validator(mode='after')
    def handle_db_directory(self):
        if isinstance(self.db_path, Path) and self.db_path.is_dir():
            self.db_path = self.db_path.joinpath(f'{self.name}.sqlite')
        return self



from collections import defaultdict
from enum import Enum
from typing import Callable, Dict, List, Optional, Sequence

from avtdl.core.interfaces import Action, ActionEntity, ActorConfig, Monitor, MonitorEntity, Record, RuntimeContext
from avtdl.core.plugins import Plugins


class QuitMode(str, Enum):
    single = 'single'
    gather = 'gather'


@Plugins.register('utils.quit', Plugins.kind.ACTOR_CONFIG)
class QuitActionConfig(ActorConfig):
    pass


@Plugins.register('utils.quit', Plugins.kind.ACTOR_ENTITY)
class QuitActionEntity(ActionEntity):
    local_counter: int = -1
    """decrement counter on every received record, quit when it reaches zero"""
    global_counter: int = -1
    """decrement counter on every received record, quit when all global counters reaches zero"""


@Plugins.register('utils.quit', Plugins.kind.ACTOR)
class QuitAction(Action):
    """
    Used for testing purposes

    Triggers avtdl shutdown when specified number of records is received.
    """
    def __init__(self, conf: QuitActionConfig, entities: Sequence[QuitActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

    def handle(self, entity: QuitActionEntity, record: Record):
        if entity.local_counter > 0:
            entity.local_counter -= 1
            self.logger.info(f'[{entity.name}] decrementing local counter, {entity.local_counter} left')
        if entity.local_counter == 0:
            self.quit(entity)

        if entity.global_counter > 0:
            entity.global_counter -= 1
            self.logger.info(f'[{entity.name}] decrementing global counter, {entity.global_counter} left for this entity')
        self.check_global_counter(entity)

    def check_global_counter(self, entity: QuitActionEntity):
        counters = [entity.global_counter for entity in self.entities.values() if entity.global_counter >= 0]
        if all(c == 0 for c in counters):
            self.quit(entity)

    def quit(self, entity: QuitActionEntity):
        msg = f'[{entity.name}] interrupting program execution'
        self.logger.info(msg)
        raise KeyboardInterrupt(msg)


Plugins.register('utils.producer', Plugins.kind.ACTOR_CONFIG)(ActorConfig)

Plugins.register('utils.producer', Plugins.kind.ACTOR_ENTITY)(MonitorEntity)


@Plugins.register('utils.producer', Plugins.kind.ACTOR)
class Producer(Monitor):
    """
    Used for testing purposes

    Produces records programmatically.
    """

    def produce(self, entity_name: str, record: Record):
        """Programmatically emit given record on behalf of own entity with given name"""
        entity = self.entities[entity_name]
        self.on_record(entity, record)


Plugins.register('utils.consumer', Plugins.kind.ACTOR_CONFIG)(ActorConfig)

Plugins.register('utils.consumer', Plugins.kind.ACTOR_ENTITY)(ActionEntity)


@Plugins.register('utils.consumer', Plugins.kind.ACTOR)
class Consumer(Action):
    """
    Used for testing purposes

    Keeps history of received records, allow registering handlers
    for records received by specific entity.
    """
    def __init__(self, conf: ActorConfig, entities: Sequence[ActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.history: Dict[str, List[Record]] = {entity.name: [] for entity in entities}
        self.callbacks: Dict[Optional[str], List[Callable]] = defaultdict(list)

    def handle(self, entity: ActionEntity, record: Record):
        if entity.name in self.history:
            self.history[entity.name].append(record)
        for callback in self.callbacks[None]:
            callback(entity, record)
        if entity.name in self.callbacks:
            for callback in self.callbacks[entity.name]:
                callback(entity, record)

    def register_callback(self, callback: Callable[[ActionEntity, Record], None], entity_name: Optional[str] = None):
        self.callbacks[entity_name].append(callback)


import logging
from functools import wraps
from pathlib import Path
from typing import Any, Dict, Generic, List, Tuple, Type, TypeVar

from pydantic import BaseModel, ConfigDict, Field, ValidationError, create_model

from avtdl.core.chain import Chain, ChainConfigSection
from avtdl.core.interfaces import Actor, RuntimeContext
from avtdl.core.loggers import LogLevel, override_loglevel, set_file_logger
from avtdl.core.plugins import Plugins
from avtdl.core.utils import strip_text


class ConfigurationError(Exception):
    """Generic exception raised if parsing config failed"""


def format_validation_error(e: ValidationError) -> str:
    msg = 'Failed to process configuration file, following errors occurred: '
    errors = []
    for err in e.errors():
        user_input = str(err['input'])
        user_input = user_input if len(user_input) < 85 else user_input[:50] + ' [...] ' + user_input[-30:]
        location = ': '.join(str(l) for l in err['loc'])
        error_message = strip_text(err['msg'], 'Value error, ')
        error = 'error parsing "{}" in config section {}: {}'
        errors.append(error.format(user_input, location, error_message))
    return '\n    '.join([msg] + errors)


def try_parsing(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValidationError as e:
            error = format_validation_error(e)
            raise ConfigurationError(error) from e

    return wrapper


class SettingsSection(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)

    log_directory: Path = Field(default='logs', validate_default=True)
    """path to a directory where application will write log file"""
    logfile_size: int = Field(gt=0, default=1000000)
    """size of a single log file in bytes. After reaching this size the file will be replaced by a new one. Only last 10 files are kept inside the log directory"""
    logfile_level: LogLevel = LogLevel.debug
    """how detailed the output to log file is. Can be "DEBUG", "INFO", "WARNING" or "ERROR". It is recommended to keep log file loglevel set to "DEBUG" """
    loglevel_override: Dict[str, LogLevel] = {'bus': LogLevel.info, 'chain': LogLevel.info,
                                              'actor.request': LogLevel.info}
    """allows to overwrite loglevel of a specific logger. Used to prevent a single talkative logger from filling up the log file"""
    port: int = Field(gt=0, le=65535, default=8080)
    """web-interface port"""
    host: str = 'localhost'
    """web-interface host, typically "127.0.0.1", "0.0.0.0" or the machine external IP"""


def configure_loggers(settings: SettingsSection):
    override_loglevel(settings.loglevel_override)
    set_file_logger(path=settings.log_directory, max_size=settings.logfile_size, level=settings.logfile_level)


class ActorConfigSection(BaseModel):
    config: dict = {}
    defaults: dict = {}
    entities: List[dict]


class Config(BaseModel):
    model_config = ConfigDict(extra='forbid')

    settings: SettingsSection = SettingsSection()
    actors: Dict[str, ActorConfigSection]
    chains: Dict[str, ChainConfigSection]


TConfig = TypeVar('TConfig')
TEntity = TypeVar('TEntity')


class SpecificActorConfigSection(BaseModel, Generic[TConfig, TEntity]):
    config: TConfig
    entities: List[TEntity]


class ActorParser:

    @staticmethod
    def flatten_actor_section(name: str, section: ActorConfigSection) -> ActorConfigSection:
        config = {**section.config, **{'name': name}, **{'defaults': section.defaults}}
        data: Dict[str, Any] = {'name': name, 'config': config, 'entities': []}
        for entity in section.entities:
            data['entities'].append({**section.defaults, **entity})
        return ActorConfigSection(**data)

    @staticmethod
    def load_actors_plugins_model(actor_section: dict) -> Dict[str, SpecificActorConfigSection]:
        actors_models: Dict[str, Any] = {}
        for name, section in actor_section.items():
            _, ConfigFactory, EntityFactory = Plugins.get_actor_factories(name)
            model = SpecificActorConfigSection[ConfigFactory, EntityFactory]
            actors_models[name] = (model, ...)
        actors_section_model = create_model('SpecificActors', **actors_models)
        return actors_section_model

    @classmethod
    def create_actors(cls, config_section: 'SpecificActors', ctx: RuntimeContext) -> Dict[str, Type]:
        actors = {}
        for name, actor_section in config_section:
            ActorFactory, _, _ = Plugins.get_actor_factories(name)
            actors[name] = ActorFactory(actor_section.config, actor_section.entities, ctx)
        return actors

    @classmethod
    def serialize_actor(cls, actor: Actor) -> ActorConfigSection:
        actor_config = actor.conf.model_dump()
        defaults = actor.conf.defaults
        entities = [entity.model_dump() for entity in actor.entities.values()]
        for entity in entities:
            for field, value in defaults.items():
                if field in entity and entity[field] == value:
                    entity.pop(field)
        section = ActorConfigSection(config=actor_config, defaults=defaults, entities=entities)
        return section


class ConfigParser:

    @staticmethod
    def flatten_config(config: Config) -> Config:
        conf = config.model_dump()
        actors_section: Dict[str, ActorConfigSection] = {}
        for name, section in config.actors.items():
            actors_section[name] = ActorParser.flatten_actor_section(name, section)
        conf['actors'] = actors_section
        return Config(**conf)

    @staticmethod
    def load_models(config: Config) -> Type['SpecificConfig']:
        actors_model = ActorParser.load_actors_plugins_model(config.actors)
        SpecificConfigModel = create_model('SpecificConfig',
                                           actors=(actors_model, ...),
                                           chains=(Dict[str, ChainConfigSection], ...)
                                           )
        return SpecificConfigModel

    @classmethod
    def create_chains(cls, chains_section: Dict[str, ChainConfigSection], ctx: RuntimeContext) -> Dict[str, Chain]:
        chains = {}
        for name, chain_config in chains_section.items():
            chains[name] = Chain(name, chain_config, ctx)
        return chains

    @classmethod
    @try_parsing
    def parse(cls, conf: dict, ctx: RuntimeContext) -> Tuple[SettingsSection, Dict[str, Any], Dict[str, Chain]]:
        # do basic structural validation of config file
        config = Config(**conf)

        configure_loggers(config.settings)
        Plugins.load()

        # after that entities transformation and specific plugins validation can be safely performed
        flatted_conf = cls.flatten_config(config)
        SpecificConfig = cls.load_models(config)
        specific_config = SpecificConfig(**flatted_conf.model_dump())

        actors = ActorParser.create_actors(specific_config.actors, ctx)
        chains = ConfigParser.create_chains(specific_config.chains, ctx)

        return config.settings, actors, chains

    @classmethod
    def serialize(cls, settings: SettingsSection, actors: Dict[str, Actor], chains: Dict[str, Chain]) -> Config:
        actors_section = {name: ActorParser.serialize_actor(actor) for name, actor in actors.items()}
        chains_section = {name: chain.conf for name, chain in chains.items()}
        config = Config(settings=settings, actors=actors_section, chains=chains_section)
        return config

    @classmethod
    @try_parsing
    def validate(cls, conf: dict) -> 'SpecificConfig':
        """try parsing object, raise ConfigurationError on failure"""
        config = Config(**conf)
        flatted_conf = cls.flatten_config(config)
        SpecificConfig = cls.load_models(config)
        return SpecificConfig(**flatted_conf.model_dump())


def config_sancheck(actors, chains):
    """check for possible non-fatal misconfiguration and issue a warning"""
    for chain_name, chain_instance in chains.items():
        for actor_name, entities in chain_instance.conf:
            actor = actors.get(actor_name)
            if actor is None:
                logging.warning(
                    f'chain "{chain_name}" references actor "{actor_name}, absent in "Actors" section. It might be a typo in the chain configuration')
                continue
            orphans = set(entities) - actor.entities.keys()
            for orphan in orphans:
                logging.warning(
                    f'chain "{chain_name}" references "{actor_name}: {orphan}", but actor "{actor_name}" has no "{orphan}" entity. It might be a typo in the chain conf configuration')



import datetime

import pytest

from avtdl.core.interfaces import RuntimeContext, TextRecord
from avtdl.plugins.twitter.extractors import TwitterRecord
from avtdl.plugins.twitter.filters import TwitterFilter, TwitterFilterConfig, TwitterFilterEntity


@pytest.fixture()
def text_record():
    return TextRecord(text='test')


@pytest.fixture()
def config():
    return TwitterFilterConfig(name='test')


@pytest.fixture()
def empty_entity():
    return TwitterFilterEntity(name='test')


@pytest.fixture()
def regular_tweet():
    return TwitterRecord(
        uid='0',
        url='https://twitter.com/alice/0',
        author='Alice',
        username='alice',
        published=datetime.datetime.fromisoformat('2022-01-01 01:00'),
        text='test tweet'
    )


@pytest.fixture()
def retweet(regular_tweet):
    return TwitterRecord(
        uid='0',
        url='https://twitter.com/alice/0',
        author='Bob',
        username='bob',
        published=datetime.datetime.fromisoformat('2022-01-01 02:00'),
        text='test tweet',
        retweet=regular_tweet
    )


@pytest.fixture()
def reply(regular_tweet):
    return TwitterRecord(
        uid='1',
        url='https://twitter.com/eve/1',
        author='Eve',
        username='eve',
        published=datetime.datetime.fromisoformat('2022-01-01 03:00'),
        text='test reply',
        replying_to_username='alice'
    )


@pytest.fixture()
def ctx():
    return RuntimeContext.create()


class TestPassthrough:

    @staticmethod
    def test_passthrough_pick(config, text_record, ctx):
        entity = TwitterFilterEntity(name='test', regular_tweet=True, reversed=False)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, text_record) == text_record

    @staticmethod
    def test_passthrough_drop(config, text_record, ctx):
        entity = TwitterFilterEntity(name='test', regular_tweet=True, reversed=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, text_record) == text_record


class TestStraight:

    @staticmethod
    def test_pick_single_condition(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', regular_tweet=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) == regular_tweet

    @staticmethod
    def test_drop_single_condition(config, retweet, ctx):
        entity = TwitterFilterEntity(name='test', regular_tweet=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, retweet) is None

    @staticmethod
    def test_pick_multiple_conditions(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', regular_tweet=True, quote=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) == regular_tweet

    @staticmethod
    def test_pick_username(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', username='alice')
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) == regular_tweet

    @staticmethod
    def test_pick_author_name(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', author='Alice')
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) == regular_tweet

    @staticmethod
    def test_drop_retweet_name(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', author='Bob')
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) is None


class TestReversed:

    @staticmethod
    def test_pick_single_condition(config, reply, ctx):
        entity = TwitterFilterEntity(name='test', regular_tweet=True, reversed=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, reply) is reply

    @staticmethod
    def test_drop_single_condition(config, retweet, ctx):
        entity = TwitterFilterEntity(name='test', retweet=True, reversed=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, retweet) is None

    @staticmethod
    def test_pick_multiple_conditions(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', reply=True, retweet=True, quote=True, reversed=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) is regular_tweet

    @staticmethod
    def test_drop_multiple_conditions(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', regular_tweet=True, quote=True, reversed=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) is None

    @staticmethod
    def test_drop_username(config, regular_tweet, ctx):
        entity = TwitterFilterEntity(name='test', username='alice', reversed=True)
        filtr = TwitterFilter(config, [entity], ctx)

        assert filtr.match(entity, regular_tweet) is None



import asyncio
import os
from pathlib import Path
from typing import Optional

import pytest

from avtdl.core.interfaces import ActorConfig, RuntimeContext
from avtdl.plugins.file.text_file import FileMonitor, FileMonitorEntity


def add_to_file(text: str, path: Path, encoding: Optional[str] = None):
    with open(path, 'a+t', encoding=encoding) as fp:
        fp.write(text)


@pytest.fixture()
def params():
    """provides default value of no additional parameters, replaced in parametrized classes"""
    return {}


@pytest.fixture()
def entity(tmp_path, params) -> FileMonitorEntity:
    entity_params = {'name': 'test_entity', 'path': tmp_path / 'data.txt'}
    entity_params.update(params)
    entity = FileMonitorEntity.model_validate(entity_params)
    return entity


@pytest.fixture()
def monitor(entity) -> FileMonitor:
    ctx = RuntimeContext.create()
    monitor = FileMonitor(ActorConfig(name='text_monitor'), [entity], ctx)
    return monitor


@pytest.mark.asyncio
async def test_no_change_on_update(entity, monitor):
    data = 'Line 1\nLine 2\n'
    add_to_file(data, entity.path, entity.encoding)

    _ = await monitor.get_new_records(entity)
    records = await monitor.get_new_records(entity)

    output = [str(record) for record in records]
    assert output == []


class TestFirstUpdate:
    testcases = {
        'default entity settings':
        (
            {},
            'Line 1\nLine 2\n',
            ['Line 1\nLine 2']
        ),
        'split_lines splits on newline by default':
        (
            {'split_lines': True},
            'Line 1\nLine 2\n',
            ['Line 1', 'Line 2']
        ),
        'split_lines with record_start and record_end provided':
        (
            {'split_lines': True, 'record_start': 'Line', 'record_end': '$'},
            '[LOG] Line 1a Line 1b\n[LOG] Unrelated record\n[LOG] Line 2',
            ['Line 1a Line 1b', 'Line 2']
        ),
        'split_lines with record_start including entire record and record_end being "match all" pattern':
        (
            {'split_lines': True, 'record_start': 'Line [a-z0-9]+', 'record_end': ''},
            '[LOG] Line 1a Line 1b\n[LOG] Line 2',
            ['Line 1a', 'Line 1b', 'Line 2']
        ),
        'with quiet_start text is discarded on first update':
        (
            {'quiet_start': True},
            'Line 1\nLine 2\n',
            []
        )
    }

    @staticmethod
    @pytest.mark.asyncio
    @pytest.mark.parametrize('params, data, expected', testcases.values(), ids=testcases.keys())
    async def test_single_update(entity, monitor, params, data, expected):
        add_to_file(data, entity.path, entity.encoding)

        records = await monitor.get_new_records(entity)

        output = [str(record) for record in records]
        assert output == expected


class TestSecondUpdateAppend:
    testcases = {
        'without follow enabled entire file content gets read on second update':
        (
            {'follow': False},
            'Line 1\nLine 2\n', 'Line 3\nLine 4\n',
            ['Line 1\nLine 2\nLine 3\nLine 4']
        ),
        'follow with second update shorter than first':
        (
            {'follow': True},
            'Line 1\nLine 2\nLine 3\n', 'Line 4\n',
            ['Line 4']
        ),
        'follow with first update shorter than second':
        (
            {'follow': True},
            'Line 1\n', 'Line 2\nLine 3\nLine 4\n',
            ['Line 2\nLine 3\nLine 4']
        ),
        'partial record stored in buffer gets merged when continuation appended':
        (
            {'follow': True, 'split_lines': True, 'record_start': r'\[LOG\]', 'record_end': r'\.'},
            '[LOG] Line 1.\n[LOG] Line 2a\n', 'Line 2b\nLine2c.\nUnrelated line.\n[LOG] Line 3.',
            ['[LOG] Line 2a\nLine 2b\nLine2c.', '[LOG] Line 3.']
        )
    }

    @staticmethod
    @pytest.mark.asyncio
    @pytest.mark.parametrize('params, data1, data2, expected', testcases.values(), ids=testcases.keys())
    async def test_append(entity, monitor, params, data1, data2, expected):
        add_to_file(data1, entity.path, entity.encoding)
        _ = await monitor.get_new_records(entity)

        await asyncio.sleep(0.01)

        add_to_file(data2, entity.path, entity.encoding)
        records = await monitor.get_new_records(entity)

        output = [str(record) for record in records]
        assert output == expected


class TestSecondUpdateRotate:
    testcases = {
        'old file content gets cut on rotate':
        (
            {'follow': False},
            'Line 1\nLine 2\n', 'Line 3\nLine 4\n',
            ['Line 3\nLine 4']
        ),
        'follow with second update shorter than first':
        (
            {'follow': True},
            'Line 1\nLine 2\nLine 3\n', 'Line 4\n',
            ['Line 4']
        ),
        'follow with first update shorter than second':
        (
            {'follow': True},
            'Line 1\n', 'Line 2\nLine 3\nLine 4\n',
            ['Line 2\nLine 3\nLine 4']
        ),
        'partial record stored in buffer is discarded on rotate':
        (
            {'follow': True, 'split_lines': True, 'record_start': r'\[LOG\]', 'record_end': r'\.'},
            '[LOG] Line 1.\n[LOG] Line 2a\n', 'Line 2b\nLine2c.\nUnrelated line.\n[LOG] Line 3.',
            ['[LOG] Line 3.']
        ),
        "quiet_start doesn't get carried over to rotated file":
        (
            {'quiet_start': True},
            '', 'Line 3\nLine 4\n',
            ['Line 3\nLine 4']
        )
    }

    @staticmethod
    @pytest.mark.asyncio
    @pytest.mark.parametrize('params, data1, data2, expected', testcases.values(), ids=testcases.keys())
    async def test_rotate(entity, monitor, params, data1, data2, expected):
        add_to_file(data1, entity.path, entity.encoding)
        _ = await monitor.get_new_records(entity)

        os.unlink(entity.path)
        await asyncio.sleep(0.01)

        add_to_file(data2, entity.path, entity.encoding)
        records = await monitor.get_new_records(entity)

        output = [str(record) for record in records]
        assert output == expected


class TestEntityValidators:
    """makes sure constructing entity only raises when expected"""

    @staticmethod
    def test_defaults():
        _ = FileMonitorEntity(name='test', path=Path('dummy'))

    @staticmethod
    def test_valid_regexp():
        _ = FileMonitorEntity(name='test', path=Path('dummy'), record_start='^.*$')

    @staticmethod
    def test_invalid_start_regexp():
        with pytest.raises(ValueError):
            _ = FileMonitorEntity(name='test', path=Path('dummy'), record_start='*invalid regexp*')

    @staticmethod
    def test_invalid_end_regexp():
        with pytest.raises(ValueError):
            _ = FileMonitorEntity(name='test', path=Path('dummy'), record_end='*invalid regexp*')



from typing import Optional, Sequence

from avtdl.core.interfaces import Filter, FilterEntity, Record, RuntimeContext
from avtdl.core.plugins import Plugins
from avtdl.plugins.filters.filters import EmptyFilterConfig
from avtdl.plugins.twitter.extractors import TwitterRecord


@Plugins.register('filter.twitter', Plugins.kind.ACTOR_CONFIG)
class TwitterFilterConfig(EmptyFilterConfig):
    pass


@Plugins.register('filter.twitter', Plugins.kind.ACTOR_ENTITY)
class TwitterFilterEntity(FilterEntity):
    retweet: bool = False
    """match retweets"""
    reply: bool = False
    """match replies"""
    quote: bool = False
    """match quotes"""
    regular_tweet: bool = False
    """match regular tweets that are not a retweet, reply or quote"""
    author: Optional[str] = None
    """match if a given string is a part of the name of the author of the tweet"""
    username: Optional[str] = None
    """match if a given string is a part of tweet author's username (without the "@" symbol)"""
    reversed: bool = False
    """drop record instead of letting it through if any of the properties matches"""


@Plugins.register('filter.twitter', Plugins.kind.ACTOR)
class TwitterFilter(Filter):
    """
    Filter `TwitterRecord` with specified properties

    Lets through `TwitterRecord` if it matches any of specified criteria.
    Enabling `reversed` setting reverses the conditions, making record match
    if none of the criteria apply.

    All records from other sources pass through without filtering.
    """

    def __init__(self, config: TwitterFilterConfig, entities: Sequence[TwitterFilterEntity], ctx: RuntimeContext):
        super().__init__(config, entities, ctx)

    def match(self, entity: FilterEntity, record: Record) -> Optional[Record]:
        assert isinstance(entity, TwitterFilterEntity)
        if not isinstance(record, TwitterRecord):
            self.logger.debug(f'[{entity.name}] record is not a TwitterRecord, letting through: {record!r}')
            return record
        matches = entity_matches(record, entity)
        if (matches and not entity.reversed) or (not matches and entity.reversed):
            return record
        else:
            return None


def is_retweet(record: TwitterRecord) -> bool:
    return record.retweet is not None


def is_reply(record: TwitterRecord) -> bool:
    return record.replying_to_username is not None


def is_quote(record: TwitterRecord) -> bool:
    return record.quote is not None


def is_regular_tweet(record: TwitterRecord) -> bool:
    return all([x is None for x in [record.replying_to_username, record.retweet, record.quote]])


def author_matches(record: TwitterRecord, name: str) -> bool:
    return record.author.find(name) > -1


def username_matches(record: TwitterRecord, name: str) -> bool:
    return record.username.find(name) > -1


def entity_matches(record: TwitterRecord, entity: TwitterFilterEntity) -> bool:
    return any([
        entity.retweet and is_retweet(record),
        entity.reply and is_reply(record),
        entity.quote and is_quote(record),
        entity.regular_tweet and is_regular_tweet(record),
        entity.author and author_matches(record, entity.author),
        entity.username and username_matches(record, entity.username)
    ])



import asyncio
import datetime
import json
import urllib.parse
from enum import Enum
from typing import Callable, Dict, Optional, Sequence

import aiohttp
from pydantic import FilePath

from avtdl.core import utils
from avtdl.core.db import BaseDbConfig, RecordDB
from avtdl.core.interfaces import Action, ActionEntity, Record, RuntimeContext
from avtdl.core.plugins import Plugins
from avtdl.core.utils import SessionStorage, find_matching_field_value
from avtdl.plugins.twitter.endpoints import AudioSpaceEndpoint, LiveStreamEndpoint
from avtdl.plugins.twitter.extractors import TwitterSpaceRecord, find_space_id, parse_media_url, parse_space, \
    space_url_by_id

Plugins.register('twitter.space', Plugins.kind.ASSOCIATED_RECORD)(TwitterSpaceRecord)

SPACE_URL_PATTERN = '/i/spaces/'


@Plugins.register('twitter.space', Plugins.kind.ACTOR_CONFIG)
class TwitterSpaceConfig(BaseDbConfig):
    pass


@Plugins.register('twitter.space', Plugins.kind.ACTOR_ENTITY)
class TwitterSpaceEntity(ActionEntity):
    cookies_file: FilePath
    """path to a text file containing cookies in Netscape format"""
    url: str = 'https://twitter.com'
    """Twitter domain name"""
    emit_immediately: bool = True
    """whether record should be produced immediately, regardless of media_url presence.
    When disabled, only records with valid media_url are produced, depending on the following settings"""
    emit_on_live: bool = True
    """if enabled, a record is produced when livestream media_url becomes available"""
    emit_on_archive: bool = True
    """if enabled, a record is produced when archive media_url becomes available"""
    emit_on_end: bool = False
    """if enabled, a record is produced when the space ends, even if there is no archive"""


@Plugins.register('twitter.space', Plugins.kind.ACTOR)
class TwitterSpace(Action):
    """
    Retrieve Twitter Space metadata from tweet

    Take a record (normally coming from a Twitter monitor), check if it
    has a link to a Twitter Space, and if so try to retrieve
    additional information on the space, such as title and start time.

    Produces a TwitterSpaceRecord if currently processed record
    contains a link to a Space and the metadata was retrieved successfully.

    It is possible to produce additional records with updated metadata at the
    beginning and/or at the end of the space by toggling the `emit_*` settings.
    However, a single state change should only produce one record. For example,
    if a space has already ended before the first update, only a single record is
    produced with all `emit_*` options enabled.
    """

    CHECK_UPCOMING_DELAY = 30.0
    CHECK_LIVE_DELAY = 300.0
    CHECK_ENDED_DELAY = 30.0
    MAX_SEQUENTIAL_CHECKS = 100

    def __init__(self, conf: TwitterSpaceConfig, entities: Sequence[TwitterSpaceEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.sessions = SessionStorage(self.logger)
        self.tasks: Dict[str, asyncio.Task] = {}
        self.db = RecordDB(conf.db_path, logger=self.logger.getChild('db'))

    def handle(self, entity: TwitterSpaceEntity, record: Record):
        space_id = get_space_id(record)
        if space_id is None:
            return
        if space_id in self.tasks:
            self.logger.debug(f'[{entity.name}] task for space {space_id} is already running')
            return
        task = self.controller.create_task(self.handle_space(entity, space_id, record))
        task.add_done_callback(lambda _: self.tasks.pop(space_id))
        self.tasks[space_id] = task

    async def run(self) -> None:
        await self.sessions.ensure_closed()

    async def handle_space(self, entity: TwitterSpaceEntity, space_id: str, source_record: Record):
        should_emit_something: bool = entity.emit_immediately
        should_emit_live_url: bool = entity.emit_on_live
        should_emit_replay_url: bool = entity.emit_on_archive
        should_emit_on_end: bool = entity.emit_on_end

        def handle_update_result(space: TwitterSpaceRecord) -> bool:
            nonlocal should_emit_something
            nonlocal should_emit_live_url
            nonlocal should_emit_replay_url
            nonlocal should_emit_on_end

            if space.chain is None:
                space.chain = source_record.chain
            if space.origin is None:
                space.origin = source_record.origin

            if should_emit_replay_url:
                if space.media_url is not None and StreamUrlType.is_replay(space.media_url):
                    self.on_record(entity, space)
                    should_emit_replay_url = False
                    should_emit_on_end = False
                    should_emit_live_url = False
                    should_emit_something = False
                    self.logger.debug(f'[{entity.name}] task emit_on_archive successfully completed for {space.url}')
                elif not space.recording_enabled:
                    should_emit_replay_url = False
                    self.logger.debug(f'[{entity.name}] task emit_on_archive for {space.url} is cancelled: recording is disabled')
            if should_emit_on_end:
                if space.ended is not None:
                    self.on_record(entity, space)
                    should_emit_on_end = False
                    should_emit_live_url = False
                    should_emit_something = False
                    self.logger.debug(f'[{entity.name}] task emit_on_end successfully completed for {space.url}')
            if should_emit_live_url:
                if space.media_url is not None and StreamUrlType.is_live(space.media_url):
                    self.on_record(entity, space)
                    should_emit_live_url = False
                    should_emit_something = False
                    self.logger.debug(f'[{entity.name}] task emit_on_live successfully completed for {space.url}')
                elif space.ended is not None:
                    should_emit_live_url = False
                    self.logger.debug(f'[{entity.name}] task emit_on_live for {space.url} is cancelled: space ended')
            if should_emit_something:
                self.on_record(entity, space)
                should_emit_something = False
                self.logger.debug(f'[{entity.name}] task emit_immediately completed for {space.url}')
            done = not (should_emit_something or should_emit_live_url or should_emit_on_end or should_emit_replay_url)
            if done:
                self.logger.debug(f'[{entity.name}] all tasks completed for {space.url}')
            return done

        session = self.sessions.get_session(entity.cookies_file)
        space = await self.fetch_space(session, entity, space_id)
        if space is None:
            return
        if self.db.record_exists(space, entity.name):
            self.logger.debug(f'[{entity.name}] space {space_id} has already been processed')
            return
        space.media_url = await self.wait_for_any_url(session, entity, space) or None
        self.db.store_records([space], entity.name)

        done = handle_update_result(space)
        if done:
            return

        while True:
            if SpaceState.is_upcoming(space):
                media_url = await self.wait_for_live(session, entity, space)
            elif SpaceState.is_ongoing(space):
                media_url = await self.wait_for_replay(session, entity, space)
                if media_url == '':
                    self.logger.debug(f'[{entity.name}] media url unavailable for running space {space.url}, updating space metadata to see if it ended')
            elif SpaceState.has_ended(space):
                media_url = await self.wait_for_any_url(session, entity, space)
                if media_url == '':
                    self.logger.debug(f'[{entity.name}] space {space.url} has ended at {space.ended} and media url is unavailable. The space likely does not have archive at this point')
                    break
            else:
                self.logger.warning(f'[{entity.name}] space {space.url} state is unknown, aborting. {space.model_dump()}')
                break

            old_media_url = space.media_url
            space = await self.fetch_space(session, entity, space_id) or space

            if media_url:
                space.media_url = media_url
            else:
                space.media_url = old_media_url
            self.db.store_records([space], entity.name)

            done = handle_update_result(space)
            if done:
                break

    async def wait_for_live(self, session: aiohttp.ClientSession, entity: TwitterSpaceEntity, space: TwitterSpaceRecord) -> Optional[str]:
        assert space.scheduled is not None, f'upcoming space has no scheduled: {space.model_dump()}'
        while True:  # waiting until start
            delay = (space.scheduled - datetime.datetime.now(tz=datetime.timezone.utc))
            delay_seconds = delay.total_seconds()
            if delay_seconds < self.CHECK_UPCOMING_DELAY:
                break
            self.logger.debug(f'[{entity.name}] space {space.url} starts at {space.scheduled}, sleeping for {delay}')
            await asyncio.sleep(delay_seconds)
            return None

        self.logger.debug(f'[{entity.name}] space {space.url} should start now, fetching media_url')

        def is_done(media_url: str) -> bool:
            return media_url != ''

        return await self.wait_for_media_url(session, entity, space, self.CHECK_UPCOMING_DELAY, self.MAX_SEQUENTIAL_CHECKS, is_done)

    async def wait_for_replay(self, session: aiohttp.ClientSession, entity: TwitterSpaceEntity, space: TwitterSpaceRecord):
        self.logger.debug(f'[{entity.name}] space {space.url} is ongoing since {space.started}, fetching live media_url')

        def is_done(media_url: str) -> bool:
            if media_url == '':
                return True
            elif StreamUrlType.is_live(media_url):
                return False
            elif StreamUrlType.is_replay(media_url):
                return True
            else:
                self.logger.debug(f'[{entity.name}] space {space.url} got media_url with unknown type: {media_url}')
                return True

        return await self.wait_for_media_url(session, entity, space, self.CHECK_LIVE_DELAY, self.MAX_SEQUENTIAL_CHECKS, is_done)

    async def wait_for_any_url(self, session: aiohttp.ClientSession, entity: TwitterSpaceEntity, space: TwitterSpaceRecord):
        return await self.wait_for_media_url(session, entity, space, self.CHECK_ENDED_DELAY, self.MAX_SEQUENTIAL_CHECKS, lambda _: True)

    async def wait_for_media_url(self, session: aiohttp.ClientSession, entity: TwitterSpaceEntity, space: TwitterSpaceRecord,
                                 retry_delay: float, max_attempts: int, is_done: Callable[[str], bool]) -> Optional[str]:
        base_retry_delay = retry_delay
        for attempt in range(max_attempts):
            media_url = await self.fetch_media_url(session, entity, space)
            if media_url is None:
                retry_delay = utils.Delay.get_next(retry_delay)
                self.logger.debug(f'[{entity.name}] {space.url} got no media_url on {attempt} attempt, retry after {retry_delay}')
            elif is_done(media_url):
                return media_url
            else:
                retry_delay = base_retry_delay
                self.logger.debug(f'[{entity.name}] {space.url} got no media_url of expected type on {attempt} attempt, retry after {retry_delay}')
            await asyncio.sleep(retry_delay)

        self.logger.debug(f'[{entity.name}] failed to fetch media_url of expected type after {max_attempts} attempts for {space.url}')
        return None

    async def fetch_space(self, session: aiohttp.ClientSession, entity: TwitterSpaceEntity, space_id: str) -> Optional[TwitterSpaceRecord]:
        self.logger.debug(f'[{entity.name}] fetching metadata for space {space_url_by_id(space_id)}')
        text = await AudioSpaceEndpoint.request(self.logger, session, entity.url, session.cookie_jar, space_id)
        if text is None:
            self.logger.warning(f'[{entity.name}] failed to retrieve metadata for {space_url_by_id(space_id)}')
            return None
        try:
            data = json.loads(text)
        except json.JSONDecodeError as e:
            self.logger.warning(f'[{entity.name}] failed to parse space {space_url_by_id(space_id)}: {e}. Raw response: {text}')
            return None
        try:
            space = parse_space(data)
        except ValueError as e:
            self.logger.warning(f'[{entity.name}] failed to parse Space metadata for "{space_url_by_id(space_id)}": {e}')
            return None
        return space

    async def fetch_media_url(self, session: aiohttp.ClientSession, entity: TwitterSpaceEntity, space: TwitterSpaceRecord) -> Optional[str]:
        """
        fetch and parse data from media_url endpoint for given space.media_key
        return the url on success, empty string on 404 response, None on any other error
        """
        self.logger.debug(f'[{entity.name}] fetch media url for space {space.url}')
        r = LiveStreamEndpoint.prepare(entity.url, session.cookie_jar, space.media_key)
        try:
            response = await utils.request_raw(r.url, session, self.logger, params=r.params, headers=r.headers, retry_times=0, raise_errors=True)
            assert response is not None, 'request_raw() returned None despite raise_errors=True'
            text = await response.text()
        except Exception as e:
            if isinstance(e, aiohttp.ClientResponseError):
                if e.status == 404:
                    self.logger.debug(f'[{entity.name}] no media url for {space.url}: {e}')
                    return ''
                else:
                    self.logger.debug(f'[{entity.name}]  got code {e.status} ({e.message}) while fetching media_url for {space.url}: {e}')
            else:
                self.logger.debug(f'[{entity.name}] failed to retrieve media url for {space.url}: {e}')
            return None
        try:
            data = json.loads(text)
        except json.JSONDecodeError as e:
            self.logger.warning(f'[{entity.name}] failed to decode json response for media_url: {e}. Raw text: "{text}"')
            return None
        media_url = parse_media_url(data)
        if media_url is None:
            self.logger.debug(f'[{entity.name}] failed to parse media url for {space.url}. Raw data: "{data}"')
            return None
        return media_url


def get_space_id(record: Record) -> Optional[str]:
    field = find_matching_field_value(record, pattern=SPACE_URL_PATTERN)
    if field is None:
        return None
    space_id = find_space_id(str(field))
    return space_id


class SpaceState(str, Enum):
    UPCOMING = 'upcoming'
    ONGOING = 'ongoing'
    ENDED = 'ended'
    UNKNOWN = 'unknown'

    @classmethod
    def from_space_record(cls, record: TwitterSpaceRecord) -> 'SpaceState':
        if record.ended is not None:
            return cls.ENDED
        elif record.started is not None:
            return cls.ONGOING
        elif record.scheduled is not None:
            return cls.UPCOMING
        else:
            return cls.UNKNOWN

    @staticmethod
    def is_upcoming(space: TwitterSpaceRecord) -> bool:
        return space.scheduled is not None and space.started is None

    @staticmethod
    def is_ongoing(space: TwitterSpaceRecord) -> bool:
        return space.started is not None and space.ended is None

    @staticmethod
    def has_ended(space: TwitterSpaceRecord) -> bool:
        return space.ended is not None

    @staticmethod
    def is_unknown(space: TwitterSpaceRecord) -> bool:
        return space.scheduled is None and space.started is None and space.ended is None


class StreamUrlType(str, Enum):
    LIVE = 'live'
    REPLAY = 'replay'
    UNKNOWN = 'unknown'

    @classmethod
    def from_url(cls, url: str) -> 'StreamUrlType':
        parsed = urllib.parse.urlparse(url)
        query = urllib.parse.parse_qs(parsed.query)
        types = query.get('type', [])
        if 'replay' in types:
            return StreamUrlType.REPLAY
        if 'live' in types:
            return StreamUrlType.LIVE
        return StreamUrlType.UNKNOWN

    @classmethod
    def is_live(cls, url: str) -> bool:
        return cls.from_url(url) == cls.LIVE

    @classmethod
    def is_replay(cls, url: str) -> bool:
        return cls.from_url(url) == cls.REPLAY



import asyncio
import datetime
import json
import traceback
from abc import abstractmethod
from typing import Any, List, Optional, Sequence, Tuple

import aiohttp
from pydantic import Field, FilePath, PositiveFloat

from avtdl.core.interfaces import Record
from avtdl.core.monitors import PagedFeedMonitor, PagedFeedMonitorConfig, PagedFeedMonitorEntity
from avtdl.core.plugins import Plugins
from avtdl.plugins.twitter.endpoints import LatestTimelineEndpoint, TimelineEndpoint, TwitterEndpoint, \
    UserIDEndpoint, UserLikesEndpoint, UserTweetsEndpoint, UserTweetsRepliesEndpoint
from avtdl.plugins.twitter.extractors import TwitterRecord, extract_contents, parse_tweet

Plugins.register('twitter.user', Plugins.kind.ASSOCIATED_RECORD)(TwitterRecord)
Plugins.register('twitter.home', Plugins.kind.ASSOCIATED_RECORD)(TwitterRecord)


@Plugins.register('twitter.user', Plugins.kind.ACTOR_CONFIG)
@Plugins.register('twitter.home', Plugins.kind.ACTOR_CONFIG)
class TwitterMonitorConfig(PagedFeedMonitorConfig):
    pass


class TwitterMonitorEntity(PagedFeedMonitorEntity):
    cookies_file: FilePath
    """path to a text file containing cookies in Netscape format"""
    update_interval: PositiveFloat = 1800
    """how often the monitored url will be checked, in seconds"""
    url: str = 'https://twitter.com'
    """Twitter domain name"""
    rate_limited_until: Optional[datetime.datetime] = Field(exclude=True, default=None)
    """internal variable to persist state between updates. Used to store time when currently active rate limit expires"""
    adjust_update_interval: bool = Field(exclude=True, default=False)
    """this monitor handles adjusting interval itself, so it is disabled to make sure superclass won't overwrite it"""


class TwitterMonitor(PagedFeedMonitor):
    """Base class for concrete Twitter monitors"""

    MIN_CONTINUATION_DELAY: int = 1

    async def handle_first_page(self, entity: TwitterMonitorEntity, session: aiohttp.ClientSession) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        raw_page = await self._get_page(entity, session, continuation=None)
        if raw_page is None:
            return None, None
        records, continuation = await self._parse_entries(raw_page)
        if not records:
            continuation = None
        return records, continuation

    async def handle_next_page(self, entity: TwitterMonitorEntity, session: aiohttp.ClientSession, context: Optional[Any]) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        continuation: Optional[str] = context
        if continuation is None:
            return None, None
        raw_page = await self._get_page(entity, session, continuation)
        if raw_page is None:
            return None, None
        records, continuation = await self._parse_entries(raw_page)
        if not records:
            # for user timeline Twitter keeps responding with cursor even if there is no tweets on continuation page
            continuation = None
        return records, continuation

    @abstractmethod
    async def _get_page(self, entity: TwitterMonitorEntity, session: aiohttp.ClientSession, continuation: Optional[str]) -> Optional[str]:
        """Retrieve raw response string from endpoint"""

    async def _parse_entries(self, page: str) -> Tuple[List[TwitterRecord], Optional[str]]:
        raw_tweets, continuation = extract_contents(page)
        await asyncio.sleep(0)
        records = []
        for tweet_result in raw_tweets:
            try:
                record = parse_tweet(tweet_result)
            except Exception as e:
                self.logger.warning(f'error parsing tweet: {e}')
                self.logger.debug(f'stack trace: {traceback.format_exc()}')
                self.logger.debug(f'raw tweet_result: {tweet_result}')
            else:
                records.append(record)
        return records, continuation


@Plugins.register('twitter.home', Plugins.kind.ACTOR_ENTITY)
class TwitterHomeMonitorEntity(TwitterMonitorEntity):
    following: bool = True
    """monitor tweets from the "Following" tab instead of "For you" """


@Plugins.register('twitter.home', Plugins.kind.ACTOR)
class TwitterHomeMonitor(TwitterMonitor):
    """
    Monitor for Twitter home timeline

    Monitors tweets on Twitter Home Timeline, either the "Following"
    of the "For you" tab.

    Requires login cookies from a logged in Twitter account to work.
    """

    async def _get_page(self, entity: TwitterHomeMonitorEntity, session: aiohttp.ClientSession, continuation: Optional[str]) -> Optional[str]:
        endpoint = LatestTimelineEndpoint if entity.following else TimelineEndpoint
        data = await endpoint.request(self.logger, session, entity.url, session.cookie_jar, continuation)
        return data

@Plugins.register('twitter.user', Plugins.kind.ACTOR_ENTITY)
class TwitterUserMonitorEntity(TwitterMonitorEntity):
    user: str
    """user handle"""
    with_replies: bool = True
    """include replies by monitored user"""
    only_likes: bool = False
    """monitor tweets liked by the user instead of user's own tweets"""
    user_id: Optional[str] = Field(exclude=True, default=None)
    """internal variable to persist state between updates. Used to cache user id for monitored user"""


@Plugins.register('twitter.user', Plugins.kind.ACTOR)
class TwitterUserMonitor(TwitterMonitor):
    """
    Monitor for user tweets

    Monitors timeline of a user for new tweets, including retweets and quotes.
    Enabling `with_replies` will additionally include replies posted by the user.

    With `only_likes` option enabled tweets from the "Likes" tab are collected
    instead of user's own tweets.

    Requires login cookies from a logged in Twitter account to work.
    """

    @staticmethod
    def _pick_endpoint(entity: TwitterUserMonitorEntity) -> type[TwitterEndpoint]:
        if entity.only_likes:
            return UserLikesEndpoint
        elif entity.with_replies:
            return UserTweetsRepliesEndpoint
        else:
            return UserTweetsEndpoint

    async def _get_user_id(self, entity: TwitterUserMonitorEntity, session: aiohttp.ClientSession) -> Optional[str]:
        if entity.user_id is None:
            text = await UserIDEndpoint.request(self.logger, session, entity.url, session.cookie_jar, entity.user)
            if text is None:
                return None
            try:
                data = json.loads(text)
            except json.JSONDecodeError as e:
                self.logger.warning(f'[{entity.name}] failed to parse user_id for @{entity.user}: {e}. Raw response: {text}')
                return None
            user_id = UserIDEndpoint.get_user_id(data)
            entity.user_id = user_id
        return entity.user_id

    async def _get_page(self, entity: TwitterUserMonitorEntity, session: aiohttp.ClientSession, continuation: Optional[str]) -> Optional[str]:
        user_id = await self._get_user_id(entity, session)
        if user_id is None:
            self.logger.warning(f'failed to get user id from user handle for "{entity.user}", aborting update')
            return None
        endpoint = self._pick_endpoint(entity)
        data = await endpoint.request(self.logger, session, entity.url, session.cookie_jar, user_id, continuation)
        return data



import json
import re
from typing import Dict, Tuple

import pytest

from avtdl.core.yaml import CommentedData, merge_data, yaml_dump, yaml_load


def merge_yaml(base: str, data_json: str) -> CommentedData:
    base = yaml_load(base)
    data = json.loads(data_json)
    merged = merge_data(base, data)
    return merged


testcases: Dict[str, Tuple[str, str, str]] = {
    'comments preserved': (
        """
        # top comment
        key:
        # mapping comment
          - "0"
          - "1" # list comment
          - 2""",
        '{"key": ["0", "1", 3]}',

        """
        # top comment
        key:
        # mapping comment
          - "0"
          - "1"
          - 3""",
    ),
    'parenthesses preserved': (
        """top:
          - "a"
          - 'b'
          - 'c'""",
        '{"top": ["a", "b", "X"]}',

        """top:
          - "a"
          - 'b'
          - X""",
    ),
    'dict update': (
        """top:
          a: 1
          b: 2
          c: 3""",
        '{"top": {"o": 0, "a": 1, "c": 3, "d": 4 }}',

        """top:
          o: 0
          a: 1
          c: 3
          d: 4""",
    ),
}


@pytest.mark.parametrize('base_yaml, json_data, expected', testcases.values(), ids=testcases.keys())
def test_top_level(base_yaml, json_data, expected):
    merged = merge_yaml(base_yaml, json_data)
    result = yaml_dump(merged)
    assert re.sub(r'[\n ]', '', result) == re.sub(r'[\n ]', '', expected)
